"""TeamCity utility that wraps TeamCity build metadata.
TeamCity server predefined build parameters can be found here:
https://www.jetbrains.com/help/teamcity/predefined-build-parameters.html
"""
import os
from typing import Optional

__all__ = ["is_running_in_teamcity", "get_build_conf_name"]


def is_running_in_teamcity() -> bool:
    return get_build_conf_name() is not None


def get_build_conf_name() -> Optional[str]:
    return os.environ.get("TEAMCITY_BUILDCONF_NAME", None)



import random

import numpy as np

functions_str = [
    # 'count',
    "sum",
    "mean",
    "median",
    "min",
    "max",
    # 'prod',
    "var",
    "quantile",
    # 'cumsum',
    # 'cumprod',
    # 'cummax',
    # 'cummin'
]

# dictionary of {func_name : (rt_cat_kwargs, pd_gb_kwargs)}
functions_with_kwargs = {"quantile": (", q=0.33", "q=0.33, interpolation='midpoint'")}

VAL_COLUMNS_MAX = 9
VAL_COLUMNS_MIN = 1
VAL_COLUMNS_INC = 1

SYMBOL_RATIO_MAX = 0.5
SYMBOL_RATIO_MIN = 0.1
SYMBOL_RATIO_INC = 0.15

SIZE_DEFAULT = 100  ## to ensure that multithreading is enabled


def isNaN(num):
    return num != num


def safe_assert(ary1, ary2):

    if len(ary1) != len(ary2):
        print("mismatch lengths-", len(ary1), len(ary2))
        assert len(ary1) == len(ary2)
    epsilon = 0.00000005  ##diference must be in range of this

    for a, b in zip(ary1, ary2):
        assert abs(a - b) < epsilon or (isNaN(a) and isNaN(b))


def remove_nan(ary):
    return [x for x in ary if not x == 0 and not isNaN(x)]


class categorical_parameters:
    def __init__(self):
        self.val_cols = VAL_COLUMNS_MIN
        self.symbol_ratio = SYMBOL_RATIO_MIN
        self.numb_keys = int(SIZE_DEFAULT * self.symbol_ratio)
        self.aggs = ""
        self.bin_ids = []
        self.aggs_id = 0
        self.kwargs = ("", "")

        self.update("aggs")

    def update(self, parameter=None, increment=None):
        def update_aggs(self):
            self.aggs = functions_str[self.aggs_id % len(functions_str)]
            if self.aggs in functions_with_kwargs:
                self.kwargs = functions_with_kwargs[self.aggs]
            else:
                self.kwargs = ("", "")
            self.aggs_id += 1

        def update_symbs(self):
            inc = SYMBOL_RATIO_INC if increment is None else increment

            self.symbol_ratio += inc
            value = self.symbol_ratio
            self.symbol_ratio %= SYMBOL_RATIO_MAX
            if self.symbol_ratio < SYMBOL_RATIO_MIN:
                self.symbol_ratio = SYMBOL_RATIO_MIN

            return value != self.symbol_ratio

        def update_val_cols(self):
            inc = VAL_COLUMNS_INC if increment is None else increment

            self.val_cols += inc
            value = self.val_cols
            self.val_cols %= VAL_COLUMNS_MAX
            self.val_cols = self.val_cols if self.val_cols > VAL_COLUMNS_MIN else VAL_COLUMNS_MIN
            return value != self.val_cols

        if parameter is None:
            if update_val_cols(self):
                update_symbs(self)
            update_aggs(self)
        else:
            switch = {
                "aggs": update_aggs,
                "syms": update_symbs,
                "vals": update_val_cols,
            }
            switch[parameter](self)


class categorical_base:
    def __init__(self, val_cols, symbol_ratio, aggs):

        self.val_cols = val_cols
        self.symbol_ratio = symbol_ratio
        self.numb_keys = int(SIZE_DEFAULT * self.symbol_ratio)
        self.aggs = aggs
        self.generate_bin_data()
        self.data = self.generate_dummy_data()

    def generate_bin_data(self):
        self.numb_keys = int(SIZE_DEFAULT * self.symbol_ratio)
        self.keys = ["label" + str(x) for x in range(0, self.numb_keys)]
        self.bin_ids = [random.randint(0, self.numb_keys - 1) for x in range(0, SIZE_DEFAULT)]

    import numpy as np

    def generate_dummy_data(self):
        col_names = ["bin" + char.upper() for char in "abcdefghijklmnopqrstuviwxyz"]
        col_names = col_names[0 : self.val_cols]

        data = {}

        for name in col_names:
            data[name] = [0.5 + x for x in np.random.rand(SIZE_DEFAULT)]

        # self.GB_KEYS = self.bin_ids
        # data['KEYS'] = self.bin_ids

        return data



"""Test around scalar constructors and scalar methods."""
import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_warns

import riptable as rt


class TestScalarConstructor(object):
    # Type-coercion from strings test cases adapted from numpy/core/tests/test_scalar_ctors.py.
    # https://github.com/numpy/numpy/blob/c31cc36a8a814ed4844a2a553454185601914a5a/numpy/core/tests/test_scalar_ctors.py
    @pytest.mark.parametrize(
        "scalar_ctor, numeric_string",
        [
            # simple numeric string
            ("single", "1.234"),
            ("double", "1.234"),
            ("longdouble", "1.234"),
            # numeric string with overflow overflow; expect inf value
            ("half", "1e10000"),
            ("single", "1e10000"),
            ("double", "1e10000"),
            ("longdouble", "1e10000"),
            ("longdouble", "-1e10000"),
        ],
    )
    def test_floating(self, scalar_ctor, numeric_string):
        rt_value = getattr(rt, scalar_ctor)(numeric_string)
        np_value = getattr(np, scalar_ctor)(numeric_string)
        assert_almost_equal(rt_value, np_value)

    @pytest.mark.parametrize(
        "scalar_ctor, numeric_string",
        [
            ("longdouble", "1e10000"),
            ("longdouble", "-1e10000"),
        ],
    )
    def test_overflow_warning(self, scalar_ctor, numeric_string):
        assert_warns(RuntimeWarning, getattr(np, scalar_ctor), numeric_string)



# TODO add test cases for half, single, and longdouble functions.
# TODO add test cases for equivalent dtypes: bool_, bytes_, str_, int0, and uint0.
# TODO add test cases for mask operations: mask_or, mask_and, mask_xor, mask_andnot, mask_ori, mask_andi, mask_xori, and mask_andnoti.
# TODO fold all int, uint and float equivalence tests into one method that uses pytest parameters
import builtins
import operator
import random
from typing import List

import hypothesis
import numpy as np
import pytest
from hypothesis import HealthCheck, assume, event, example, given
from hypothesis.extra.numpy import (
    arrays,
    basic_indices,
    boolean_dtypes,
    floating_dtypes,
    integer_dtypes,
)
from hypothesis.strategies import (
    booleans,
    data,
    floats,
    integers,
    just,
    lists,
    sampled_from,
    shared,
    slices,
    text,
)
from numpy.testing import assert_allclose

import riptable as rt
from riptable.rt_numpy import QUANTILE_METHOD_NP_KW
from riptable import FA, FastArray
from riptable.Utils.rt_testing import (
    _NDARRAY_FASTARRAY_COMMON_AND_DIFF,
    assert_array_equal_,
    assert_equal_,
)
from riptable.Utils.teamcity_helper import is_running_in_teamcity

# riptable custom Hypothesis strategies
from .strategies.helper_strategies import (
    _MAX_FLOAT,
    _MAX_INT,
    _MAX_SHAPE_SIZE,
    _MAX_VALUE,
    floating_scalar,
    generate_array_and_axis,
    generate_array_and_where,
    generate_array_axis_and_ddof,
    generate_array_axis_and_repeats_array,
    generate_reshape_array_and_shape_strategy,
    generate_sample_test_floats,
    generate_sample_test_integers,
    generate_tuples_of_arrays,
    interpolation_data,
    ints_floats_complex_or_booleans,
    ints_floats_datetimes_and_timedeltas,
    ints_floats_or_complex_dtypes,
    ints_or_floats_dtypes,
    ints_or_floats_example,
    ndarray_shape_strategy,
    one_darray_shape_strategy,
    same_structure_ndarrays,
    start_stop_step_strategy,
)

"""
Commonalities between riptable and numpy
>> import numpy as np
>> import riptable as rt
>> sorted(list(set(dir(rt)).intersection(set(dir(np)))))
>> ['__builtins__', '__cached__', '__doc__', '__file__', '__loader__', '__name__', '__package__', '__path__', '__spec__', '__version__', 'abs', 'absolute', 'all', 'any', 'arange', 'argsort', 'bincount', 'ceil', 'concatenate', 'cumsum', 'diff', 'double', 'empty', 'empty_like', 'float32', 'float64', 'floor', 'full', 'hstack', 'int16', 'int32', 'int64', 'int8', 'interp', 'isfinite', 'isinf', 'isnan', 'lexsort', 'log', 'log10', 'max', 'maximum', 'mean', 'median', 'min', 'minimum', 'nan', 'nan_to_num', 'nanmax', 'nanmean', 'nanmedian', 'nanmin', 'nanpercentile', 'nanstd', 'nansum', 'nanvar', 'ones', 'ones_like', 'percentile', 'putmask', 'quantile', 'repeat', 'reshape', 'round', 'searchsorted', 'single', 'size', 'sort', 'std', 'sum', 'tile', 'transpose', 'trunc', 'uint16', 'uint32', 'uint64', 'uint8', 'unique', 'var', 'vstack', 'where', 'zeros', 'zeros_like']
"""


class TestRiptableNumpyEquivalency:
    @pytest.mark.xfail(reason="SOQTEST-6479 abs calls np.abs instead of rt.absolute")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(arr_and_where=generate_array_and_where(shape=ndarray_shape_strategy(), dtype=ints_or_floats_dtypes()))
    def test_abs(self, arr_and_where):
        arr, where_arr = arr_and_where

        # Test #1: Returns new array containing results; no optional parameters provided.
        rt_abs = rt.abs(arr)
        np_abs = np.abs(arr)
        assert isinstance(rt_abs, FastArray)
        assert_array_equal_(np.array(rt_abs), np_abs)

        # Test #2: Returns new array containing results; 'where' bitmask is provided.
        rt_abs = rt.abs(arr, where=where_arr)
        np_abs = np.abs(arr, where=where_arr)
        assert isinstance(rt_abs, FastArray)
        assert_array_equal_(np.array(rt_abs), np_abs)

        # Test #3: Results written to array specified in 'out' parameter; no other optional params.
        rt_abs = np.zeros_like(arr)
        np_abs = np.zeros_like(arr)
        rt_output = rt.abs(arr, out=rt_abs)
        np_output = np.abs(arr, out=np_abs)
        assert_array_equal_(np.array(rt_abs), np_abs)
        # TODO: Add assertions for rt_output and np_output -- what are they expected to return when the 'out' parameter is specified?
        # assert isinstance(rt_output, FastArray, msg="riptable.abs() did not return a FastArray")

        # Test #4: Results written to array specified in 'out' parameter; 'where' bitmask is provided.
        rt_abs = np.zeros_like(arr)
        np_abs = np.zeros_like(arr)
        rt_output = rt.abs(arr, where=where_arr, out=rt_abs)
        np_output = np.abs(arr, where=where_arr, out=np_abs)
        assert_array_equal_(np.array(rt_abs), np_abs)
        # TODO: Add assertions for rt_output and np_output -- what are they expected to return when the 'out' parameter is specified?
        # assert isinstance(rt_output, FastArray, msg="riptable.abs() did not return a FastArray")

    @pytest.mark.xfail(reason="RIP-357 discrepency between rt.absolute and np.absolute")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(arr_and_where=generate_array_and_where(shape=ndarray_shape_strategy(), dtype=ints_or_floats_dtypes()))
    def test_absolute(self, arr_and_where):
        arr, where_arr = arr_and_where

        # Test #1: Returns new array containing results; no optional parameters provided.
        rt_output = rt.absolute(arr)
        np_output = np.absolute(arr)
        assert isinstance(rt_output, FastArray)
        assert_array_equal_(np.array(rt_output), np_output)

        # Test #2: Returns new array containing results; 'where' bitmask is provided.
        rt_output = rt.absolute(arr, where=where_arr)
        np_absolute = np.absolute(arr, where=where_arr)
        assert isinstance(rt_output, FastArray)
        assert_array_equal_(np.array(rt_output), np_output)

        # Test #3: Results written to array specified in 'out' parameter; no other optional params.
        rt_inplace = np.zeros_like(arr)
        np_inplace = np.zeros_like(arr)
        rt_output = rt.absolute(arr, out=rt_inplace)
        np_output = np.absolute(arr, out=np_inplace)
        assert_array_equal_(np.array(rt_inplace), np_inplace)
        # TODO: Add assertions for rt_output and np_output -- what are they expected to return when the 'out' parameter is specified?
        # assert isinstance(rt_output, FastArray, msg="riptable.absolute() did not return a FastArray")

        # Test #4: Results written to array specified in 'out' parameter; 'where' bitmask is provided.
        rt_inplace = np.zeros_like(arr)
        np_inplace = np.zeros_like(arr)
        rt_output = rt.absolute(arr, where=where_arr, out=rt_inplace)
        np_output = np.absolute(arr, where=where_arr, out=np_inplace)
        assert_array_equal_(np.array(rt_inplace), np_inplace)
        # TODO: Add assertions for rt_output and np_output -- what are they expected to return when the 'out' parameter is specified?
        # assert isinstance(rt_output, FastArray, msg="riptable.absolute() did not return a FastArray")

    @given(
        arr_and_axis=generate_array_and_axis(shape=ndarray_shape_strategy(), dtype=ints_floats_complex_or_booleans())
    )
    def test_all(self, arr_and_axis):
        arr = arr_and_axis[0]
        axis = arr_and_axis[1]
        rt_all = rt.all(arr, axis=axis)
        np_all = np.all(arr, axis=axis)
        if axis is None:
            # If axis is None, all is performed on entire matrix, so it should have the same result as the builtin all()
            built_in_all = builtins.all(arr.flatten())
            assert rt_all == built_in_all
            assert rt_all == np_all
        else:
            # if performing all() over a certain axis, it will return an array/matrix
            assert_array_equal_(np.array(rt_all), np_all)

    @given(
        arr_and_axis=generate_array_and_axis(shape=ndarray_shape_strategy(), dtype=ints_floats_complex_or_booleans())
    )
    def test_any(self, arr_and_axis):
        arr = arr_and_axis[0]
        axis = arr_and_axis[1]
        rt_any = rt.any(arr, axis=axis)
        np_any = np.any(arr, axis=axis)
        if axis is None:
            # If axis is None, any is performed on entire matrix, so it should have the same result as the builtin any()
            built_in_any = builtins.any(arr.flatten())
            assert rt_any == built_in_any
            assert rt_any == np_any
        else:
            # if performing any() over a certain axis, it will return an array/matrix
            assert_array_equal_(np.array(rt_any), np_any)

    @hypothesis.settings(suppress_health_check=[HealthCheck.too_slow])
    @example(
        start_stop_step={
            "start": (-1000000 - 1000000j),
            "stop": (1000000 + 1000000j),
            "step": (0.0000001 + 0.0000001j),
        }
    )
    @given(start_stop_step_strategy())
    def test_arange(self, start_stop_step):
        start = start_stop_step["start"]
        stop = start_stop_step["stop"]
        step = start_stop_step["step"]

        # This block is instead of the assume(length < MAX_VALUE) so the example can be used to test large
        # numbers and complex numbers.
        min_step = abs((stop - start) / _MAX_VALUE)
        if min_step > abs(step):
            step = (step / abs(step)) * min_step

        # Compare results of arange and exception message, if one is thrown.
        rt_error_msg = ""
        np_error_msg = ""
        try:
            rt_arange = rt.arange(start, stop, step)
        except Exception as e:
            rt_arange = None
            rt_error_msg = str(e)
        try:
            np_arange = np.arange(start, stop, step)
        except Exception as e:
            np_arange = None
            np_error_msg = str(e)

        if rt_error_msg and np_error_msg:
            assert rt_error_msg == np_error_msg
        else:
            assert_array_equal_(np.array(rt_arange), np_arange)
            assert isinstance(rt_arange, FastArray)

    # TODO: Delete this test once SOQTEST-6495 is resolved. This way arange's main functionality can still be tested
    @pytest.mark.xfail(reason="SOQTEST-6495 kwargs not implemented in riptable.arange()")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(start_stop_step_strategy())
    def test_arange_kwargs(self, start_stop_step):
        start = start_stop_step["start"]
        stop = start_stop_step["stop"]
        step = start_stop_step["step"]
        # limit step size so lists don't get too long and use up too many resources
        # These can be smaller than in test_arange, because this one is just confirming the kwargs work
        max_length = 1000
        length = abs((stop - start) / step)
        # TODO directly draw from this range, instead of using hypothesis assume
        assume(-1 <= length < max_length)
        rt_error_msg = ""
        np_error_msg = ""
        try:
            rt_arange = rt.arange(start=start, stop=stop, step=step)
        except Exception as e:
            rt_arange = None
            rt_error_msg = str(e)
        try:
            np_arange = np.arange(start=start, stop=stop, step=step)
        except Exception as e:
            np_arange = None
            np_error_msg = str(e)

        if rt_error_msg and np_error_msg:
            assert rt_error_msg == np_error_msg
        else:
            assert_array_equal_(np.array(rt_arange), np_arange)
            assert isinstance(rt_arange, FastArray)

    @given(arrays(shape=ndarray_shape_strategy(), dtype=ints_floats_datetimes_and_timedeltas()))
    @pytest.mark.skip(reason="Skip if riptable defers to numpy argsort implementation.")
    def test_argsort(self, arr):
        axis_list = [None]
        axis_list.extend(list(range(-1, len(arr.shape))))
        sort_algs = ["quicksort", "mergesort", "heapsort", "stable"]

        for axis in axis_list:
            for sort_alg in sort_algs:
                rt_argsort = rt.argsort(arr, axis=axis, kind=sort_alg)
                np_argsort = np.argsort(arr, axis=axis, kind=sort_alg)
                assert isinstance(rt_argsort, FastArray)
                assert_array_equal_(np.array(rt_argsort), np_argsort)

    @pytest.mark.xfail(reason="SOQTEST-6497 riptable.bincount returns numpy array instead of FastArray")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        arr=arrays(dtype=integer_dtypes(), shape=ndarray_shape_strategy(max_rank=1)),
        use_weights=booleans(),
        min_length=integers(),
        data=data(),
    )
    def test_bincount(self, data, arr, use_weights, min_length):
        weights = None
        if use_weights:
            weights = data.draw(arrays(dtype=integer_dtypes(), shape=arr.shape))
        rt_err = np_err = rt_bincount = np_bincount = None

        try:
            rt_bincount = rt.bincount(arr, weights=weights, minlength=min_length)
        except Exception as e:
            rt_err = str(e)

        try:
            np_bincount = np.bincount(arr, weights=weights, minlength=min_length)
        except Exception as e:
            np_err = str(e)

        # TODO re-visit this implementation since if-clause will evaluate when there are no exception
        if rt_err is None or np_err is None:
            assert rt_err == np_err
        else:
            assert_array_equal_(np.array(rt_bincount), np_bincount)
            assert isinstance(rt_bincount, FastArray)

    @pytest.mark.xfail(reason="Related to SOQTEST-6478 different endiannesses not handled")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        data=data(),
        arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes(sizes=(32, 64))),
    )
    def test_ceil_array(self, data, arr):
        # TODO: Modify this to use the 'generate_array_and_where' strategy instead?
        where_arr = data.draw(arrays(shape=arr.shape, dtype=boolean_dtypes()))

        # Test #1: Returns new array containing results; no optional parameters provided.
        rt_output = rt.ceil(arr)
        np_output = np.ceil(arr)
        assert isinstance(rt_output, FastArray)
        assert_array_equal_(np.array(rt_output), np_output)

        # Test #2: Returns new array containing results; 'where' bitmask is provided.
        rt_output = rt.ceil(arr, where=where_arr)
        np_output = np.ceil(arr, where=where_arr)
        assert isinstance(rt_output, FastArray)
        assert_array_equal_(np.array(rt_output), np_output)

        # Test #3: Results written to array specified in 'out' parameter; no other optional params.
        rt_inplace = np.zeros_like(arr)
        np_inplace = np.zeros_like(arr)
        rt_output = rt.ceil(arr, out=rt_inplace)
        np_output = np.ceil(arr, out=np_inplace)
        assert_array_equal_(np.array(rt_inplace), np_inplace)
        # TODO: Add assertions for rt_output and np_output -- what are they expected to return when the 'out' parameter is specified?
        # assert isinstance(rt_output, FastArray, msg="riptable.ceil() did not return a FastArray")

        # Test #4: Results written to array specified in 'out' parameter; 'where' bitmask is provided.
        rt_inplace = np.zeros_like(arr)
        np_inplace = np.zeros_like(arr)
        rt_output = rt.ceil(arr, where=where_arr, out=rt_inplace)
        np_output = np.ceil(arr, where=where_arr, out=np_inplace)
        assert_array_equal_(np.array(rt_inplace), np_inplace)
        # TODO: Add assertions for rt_output and np_output -- what are they expected to return when the 'out' parameter is specified?
        # assert isinstance(rt_output, FastArray, msg="riptable.ceil() did not return a FastArray")

    @given(scalar=floating_scalar())
    def test_ceil_scalar(self, scalar):
        # test scalars
        rt_ceil_scalar = rt.ceil(scalar)
        np_ceil_scalar = np.ceil(scalar)
        assert_equal_(rt_ceil_scalar, np_ceil_scalar)

    # TODO fold the concatenate tests by using pytest params.
    # 20200303 N.B, rt.concatenate defers to np.concatenate.
    @hypothesis.settings(suppress_health_check=[HealthCheck.too_slow])
    @example(tuple_of_arrays=tuple())
    @given(tuple_of_arrays=generate_tuples_of_arrays(all_same_width=True))
    def test_concatenate_first_axis(self, tuple_of_arrays):
        rt_err_type = np_err_type = rt_concatenate = np_concatenate = None

        # Capture the output if evaluation succeeds, otherwise capture the error type.
        try:
            rt_concatenate = rt.concatenate(tuple_of_arrays, axis=0)
        except Exception as e:
            rt_err_type = type(e)
        try:
            np_concatenate = np.concatenate(tuple_of_arrays, axis=0)
        except Exception as e:
            np_err_type = type(e)

        # The concatenated arrays should be equal to a tolerance and if any exceptions were raised
        # they should be the same type.
        if rt_err_type and np_err_type:
            assert rt_err_type == np_err_type
        else:
            assert isinstance(rt_concatenate, FastArray)
            assert_array_equal_(np.array(rt_concatenate), np_concatenate)

    @hypothesis.settings(suppress_health_check=[HealthCheck.too_slow])
    @example(tuple_of_arrays=tuple())
    @given(tuple_of_arrays=generate_tuples_of_arrays(all_same_width=True))
    def test_concatenate_flatten(self, tuple_of_arrays):
        rt_err_type = np_err_type = rt_concatenate = np_concatenate = None

        # Capture the output if evaluation succeeds, otherwise capture the error type.
        try:
            rt_concatenate = rt.concatenate(tuple_of_arrays, axis=None)
        except Exception as e:
            rt_err_type = type(e)
        try:
            np_concatenate = np.concatenate(tuple_of_arrays, axis=None)
        except Exception as e:
            np_err_type = type(e)

        # The concatenated arrays should be equal to a tolerance and if any exceptions were raised
        # they should be the same type.
        if rt_err_type and np_err_type:
            assert rt_err_type == np_err_type, f"Exceptions should be the same type."
        else:
            assert isinstance(rt_concatenate, FastArray)
            assert_array_equal_(np.array(rt_concatenate), np_concatenate)

    @hypothesis.settings(suppress_health_check=[HealthCheck.too_slow])
    @example(tuple_of_arrays=tuple())
    @given(tuple_of_arrays=generate_tuples_of_arrays(all_same_height=True))
    def test_concatenate_second_axis(self, tuple_of_arrays):
        rt_err_type = np_err_type = rt_concatenate = np_concatenate = None

        # Capture the output if evaluation succeeds, otherwise capture the error type.
        try:
            rt_concatenate = rt.concatenate(tuple_of_arrays, axis=1)
        except ValueError as e:
            rt_err_type = ValueError
        try:
            np_concatenate = np.concatenate(tuple_of_arrays, axis=1)
        except ValueError as e:
            np_err_type = ValueError

        # The concatenated arrays should be equal to a tolerance and if any exceptions were raised
        # they should be the same type.
        if rt_err_type and np_err_type:
            assert rt_err_type == np_err_type
        else:
            assert isinstance(rt_concatenate, FastArray)
            assert_array_equal_(np.array(rt_concatenate), np_concatenate)

    @given(
        array_and_axis=generate_array_and_axis(shape=ndarray_shape_strategy(), dtype=ints_floats_or_complex_dtypes()),
        output_dtype=ints_floats_or_complex_dtypes(),
    )
    def test_cumsum(self, array_and_axis, output_dtype):
        arr, axis = array_and_axis

        output_dtype = output_dtype
        rt_cumsum = rt.cumsum(arr, axis=axis)  # 20200303 defers to numpy
        np_cumsum = np.cumsum(arr, axis=axis)
        assert isinstance(rt_cumsum, FastArray)
        assert_array_equal_(np.array(rt_cumsum), np_cumsum)

        rt_cumsum = rt.cumsum(arr, axis=axis, dtype=output_dtype)
        np_cumsum = np.cumsum(arr, axis=axis, dtype=output_dtype)
        assert isinstance(rt_cumsum, FastArray)
        assert rt_cumsum.dtype == output_dtype, f"Dtype should be the same as input array."
        assert_array_equal_(np.array(rt_cumsum), np_cumsum)

    @given(
        array_and_axis=generate_array_and_axis(
            shape=ndarray_shape_strategy(),
            dtype=ints_floats_or_complex_dtypes(),
            # If not specified, np.diff uses -1 as the default axis.
            default_axis=-1,
        ),
        # 'diff_iters': the number of differencing iterations the 'diff' function will perform.
        # This is kind of like the "divided differences" algorithm in that it's recursive, but
        # there's no division step. Specifying a large number of iterations for this is prohibitively
        # expensive and will cause the test to time out; we can test with a small-ish number of
        # iterations and still have good confidence we're covering all the important code paths.
        diff_iters=integers(min_value=0, max_value=8),
    )
    def test_diff(self, array_and_axis, diff_iters: int):
        arr, axis = array_and_axis

        # Record some events so when hypothesis reports test statistics we'll be able
        # to see roughly which parts of the search space were covered.
        event(f"dtype = {np.dtype(arr.dtype).name}")
        event(f"ndim = {len(arr.shape)}")

        # If we'll have to clamp the number of differences below for any of the array's axes,
        # record that as an event too -- so we'll know if we're running into that too often.
        if min(arr.shape) <= diff_iters:
            event("Clamped diff_iters for at least one axis of the array.")

        # We've drawn a random integer to use for the number of differencing steps to perform.
        # If the length of the current array axis is smaller, clamp it here to avoid an error.
        num_diffs = min(arr.shape[axis], diff_iters)

        # TODO parameterize over kwargs; but don't loop here, it'll be too slow.
        #  Draw the kwargs values (or None) from hypothesis strategies and let it
        #  decide how to search through the parameter space.

        rt_diff = rt.diff(
            arr, axis=axis, n=num_diffs
        )  # as of 20200303 defers to numpy (i.e. no riptable-specific implementation)
        np_diff = np.diff(arr, axis=axis, n=num_diffs)
        assert isinstance(rt_diff, FastArray)
        assert_array_equal_(np.array(rt_diff), np_diff)

    @given(examples=generate_sample_test_floats())
    def test_double(self, examples):
        for i in examples:
            rt_double = rt.double(i)
            np_double = np.double(i)
            assert_equal_(rt_double, np_double)
            assert isinstance(rt_double, type(np_double))

            rt_plus = rt_double + 1
            np_plus = np_double + 1
            assert_equal_(rt_plus, np_plus)
            assert isinstance(rt_plus, type(np_plus))

            rt_minus = rt_double - 1
            np_minus = np_double - 1
            assert_equal_(rt_minus, np_minus)
            assert isinstance(rt_minus, type(np_minus))

            rt_square = rt_double**2
            np_square = np_double**2
            assert_equal_(rt_square, np_square)
            assert isinstance(rt_square, type(np_square))

            rt_sqrt = np.sqrt(rt_double)
            np_sqrt = np.sqrt(np_double)
            assert_equal_(rt_sqrt, np_sqrt)
            isinstance(rt_sqrt, type(np_sqrt))

    @given(
        shape=ndarray_shape_strategy(),
        dtype=ints_floats_datetimes_and_timedeltas(),
        order=sampled_from(("F", "C")),
    )
    def test_empty(self, shape, dtype, order):
        # Empty does not initialize the values in the array, so it cannot be compared to a numpy array element-wise
        rt_empty = rt.empty(shape=shape, dtype=dtype, order=order)
        assert isinstance(rt_empty, FastArray)
        assert_equal_(shape, rt_empty.shape)
        assert_equal_(dtype.type, rt_empty.dtype.type)

        # 1-D arrays always use Column-order. Otherwise, use the order specified
        if len(rt_empty.shape) > 1 and min(rt_empty.shape) > 1:
            assert np.isfortran(rt_empty) == (order == "F")
        else:
            assert not np.isfortran(rt_empty)

    # TODO pull the non-subok parts so these are running tests
    @pytest.mark.xfail(reason="SOQTEST-6563 riptable does not implement subok")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        arr=arrays(shape=ndarray_shape_strategy(), dtype=ints_floats_datetimes_and_timedeltas()),
        dtype=ints_floats_or_complex_dtypes(),
        order=sampled_from(("C", "F", "K", "A")),
        subok=booleans(),
    )
    def test_empty_like(self, arr, dtype, order, subok):
        # TODO argument sweep
        rt_empty_like = rt.empty_like(arr, dtype=dtype, order=order, subok=subok)
        assert_equal_(dtype, rt_empty_like.dtype)
        assert_equal_(rt_empty_like.shape, arr.shape)

        # 1-D arrays always use Column-order. Otherwise, use the order specified
        if (
            len(rt_empty_like.shape) > 1
            and min(rt_empty_like.shape) > 1
            and (order == "F" or ((order == "A" or order == "K") and np.isfortran(arr)))
        ):
            assert np.isfortran(rt_empty_like)
        else:
            assert not np.isfortran(rt_empty_like)

        if subok:
            assert isinstance(rt_empty_like, FastArray)
        else:
            assert isinstance(rt_empty_like, np.ndarray)
            # FastArray is a subclass of np.ndarray, so also ensure it is not a FastArray
            assert not isinstance(rt_empty_like, FastArray)

    @given(examples=generate_sample_test_floats())
    def test_float32(self, examples):
        for i in examples:
            rt_float32 = rt.float32(i)
            np_float32 = np.float32(i)
            assert_equal_(rt_float32, np_float32)
            assert isinstance(rt_float32, type(np_float32))

            rt_plus = rt_float32 + 1
            np_plus = np_float32 + 1
            assert_equal_(rt_plus, np_plus)
            assert isinstance(rt_plus, type(np_plus))

            rt_minus = rt_float32 - 1
            np_minus = np_float32 - 1
            assert_equal_(rt_minus, np_minus)
            assert isinstance(rt_minus, type(np_minus))

            rt_square = rt_float32**2
            np_square = np_float32**2
            assert_equal_(rt_square, np_square)
            assert isinstance(rt_square, type(np_square))

            rt_sqrt = np.sqrt(rt_float32)
            np_sqrt = np.sqrt(np_float32)
            assert_equal_(rt_sqrt, np_sqrt)
            assert isinstance(rt_sqrt, type(np_sqrt))

    @given(examples=generate_sample_test_floats())
    def test_float64(self, examples):
        for i in examples:
            rt_float64 = rt.float64(i)
            np_float64 = np.float64(i)
            assert_equal_(rt_float64, np_float64)
            assert isinstance(rt_float64, type(np_float64))

            rt_plus = rt_float64 + 1
            np_plus = np_float64 + 1
            assert_equal_(rt_plus, np_plus)
            assert isinstance(rt_plus, type(np_plus))

            rt_minus = rt_float64 - 1
            np_minus = np_float64 - 1
            assert_equal_(rt_minus, np_minus)
            assert isinstance(rt_minus, type(np_minus))

            rt_square = rt_float64**2
            np_square = np_float64**2
            assert_equal_(rt_square, np_square)
            assert isinstance(rt_square, type(np_square))

            rt_sqrt = np.sqrt(rt_float64)
            np_sqrt = np.sqrt(np_float64)
            assert_equal_(rt_sqrt, np_sqrt)
            assert isinstance(rt_sqrt, type(np_sqrt))

    @pytest.mark.xfail(reason="Related to SOQTEST-6478 different endiannesses not handled")
    @given(
        data=data(),
        arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes(sizes=(32, 64))),
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    def test_floor_array(self, data, arr):
        # TODO: Modify this to use the 'generate_array_and_where' strategy instead?
        where_arr = data.draw(arrays(shape=arr.shape, dtype=boolean_dtypes()))

        # Test #1: Returns new array containing results; no optional parameters provided.
        rt_output = rt.floor(arr)
        np_output = np.floor(arr)
        assert isinstance(rt_output, FastArray)
        assert_array_equal_(np.array(rt_output), np_output)

        # Test #2: Returns new array containing results; 'where' bitmask is provided.
        rt_output = rt.floor(arr, where=where_arr)
        np_output = np.floor(arr, where=where_arr)
        assert isinstance(rt_output, FastArray)
        assert_array_equal_(np.array(rt_output), np_output)

        # Test #3: Results written to array specified in 'out' parameter; no other optional params.
        rt_inplace = np.zeros_like(arr)
        np_inplace = np.zeros_like(arr)
        rt_output = rt.floor(arr, out=rt_inplace)
        np_output = np.floor(arr, out=np_inplace)
        assert_array_equal_(np.array(rt_inplace), np_inplace)
        # TODO: Add assertions for rt_output and np_output -- what are they expected to return when the 'out' parameter is specified?
        # assert isinstance(rt_output, FastArray, msg="riptable.floor() did not return a FastArray")

        # Test #4: Results written to array specified in 'out' parameter; 'where' bitmask is provided.
        rt_inplace = np.zeros_like(arr)
        np_inplace = np.zeros_like(arr)
        rt_output = rt.floor(arr, where=where_arr, out=rt_inplace)
        np_output = np.floor(arr, where=where_arr, out=np_inplace)
        assert_array_equal_(np.array(rt_inplace), np_inplace)
        # TODO: Add assertions for rt_output and np_output -- what are they expected to return when the 'out' parameter is specified?
        # assert isinstance(rt_output, FastArray, msg="riptable.floor() did not return a FastArray")

    @given(scalar=floating_scalar())
    def test_floor_scalar(self, scalar):
        # test scalars
        rt_floor_scalar = rt.floor(scalar)
        np_floor_scalar = np.floor(scalar)
        assert_equal_(np_floor_scalar, rt_floor_scalar)

    @given(
        shape=ndarray_shape_strategy(),
        fill_value=ints_or_floats_example(),
        dtype=ints_or_floats_dtypes(),
        order=sampled_from(("C", "F")),
    )
    def test_full(self, shape, fill_value, dtype, order):
        rt_err_type = np_err_type = rt_full = np_full = None
        try:
            rt_full = rt.full(shape, fill_value, dtype=dtype, order=order)
        except Exception as e:
            rt_err_type = type(e)
        try:
            np_full = np.full(shape=shape, fill_value=fill_value, dtype=dtype, order=order)
        except Exception as e:
            np_err_type = type(e)

        if rt_err_type and np_err_type:
            assert rt_err_type == np_err_type
        else:
            assert_array_equal_(np.array(rt_full), np_full)
            assert isinstance(rt_full, FastArray)
            assert np.isfortran(rt_full) == np.isfortran(np_full)

    @given(
        shape=ndarray_shape_strategy(),
        fill_value=text(),
        order=sampled_from(("C", "F")),
    )
    def test_full_str(self, shape, fill_value, order):
        rt_full = rt.full(shape, fill_value, order=order)
        np_full = np.full(shape, fill_value, order=order)
        assert_array_equal_(np.array(rt_full), np_full)
        assert isinstance(rt_full, FastArray)
        assert np.isfortran(rt_full) == np.isfortran(np_full)

    @pytest.mark.xfail(
        reason="SOQTEST-6495 kwargs not implemented in riptable.full()",
        raises=ValueError,
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        shape=ndarray_shape_strategy(),
        fill_value=ints_or_floats_example(),
        dtype=ints_or_floats_dtypes(),
        order=sampled_from(("C", "F")),
    )
    def test_full_kwargs(self, shape, fill_value, dtype, order):
        rt_err_type = np_err_type = rt_full = np_full = None
        try:
            rt_full = rt.full(shape=shape, fill_value=fill_value, dtype=dtype, order=order)
        except Exception as e:
            rt_err_type = type(e)
        try:
            np_full = np.full(shape=shape, fill_value=fill_value, dtype=dtype, order=order)
        except Exception as e:
            np_err_type = type(e)

        if rt_err_type and np_err_type:
            assert rt_err_type == np_err_type
        else:
            assert_array_equal_(np.array(rt_full), np_full)
            assert isinstance(rt_full, FastArray)
            assert np.isfortran(rt_full) == np.isfortran(np_full)

    @hypothesis.settings(suppress_health_check=[HealthCheck.too_slow])
    @given(tuple_of_arrays=generate_tuples_of_arrays(all_same_height=True))
    def test_hstack(self, tuple_of_arrays):
        rt_err_type = np_err_type = rt_hstack = np_hstack = None
        try:
            rt_hstack = rt.hstack(tuple_of_arrays)
        except Exception as e:
            rt_err_type = type(e)
        try:
            np_hstack = np.hstack(tuple_of_arrays)
        except Exception as e:
            np_err_type = type(e)

        if rt_err_type and np_err_type:
            assert rt_err_type == np_err_type
        else:
            assert isinstance(rt_hstack, FastArray)
            assert_array_equal_(np.array(rt_hstack), np_hstack)

    @given(examples=generate_sample_test_integers(num_bits=8, signed=True))
    def test_int8(self, examples):
        for i in examples:
            rt_int8 = rt.int8(i)
            np_int8 = np.int8(i)
            assert_equal_(rt_int8, np_int8)
            assert isinstance(rt_int8, type(np_int8))

            rt_plus = rt_int8 + 1
            np_plus = np_int8 + 1
            assert_equal_(rt_plus, np_plus)
            assert isinstance(rt_plus, type(np_plus))

            rt_minus = rt_int8 - 1
            np_minus = np_int8 - 1
            assert_equal_(rt_minus, np_minus)
            assert isinstance(rt_minus, type(np_minus))

    @given(examples=generate_sample_test_integers(num_bits=16, signed=True))
    def test_int16(self, examples):
        for i in examples:
            rt_int16 = rt.int16(i)
            np_int16 = np.int16(i)
            assert_equal_(rt_int16, np_int16)
            assert isinstance(rt_int16, type(np_int16))

            rt_plus = rt_int16 + 1
            np_plus = np_int16 + 1
            assert_equal_(rt_plus, np_plus)
            assert isinstance(rt_plus, type(np_plus))

            rt_minus = rt_int16 - 1
            np_minus = np_int16 - 1
            assert_equal_(rt_minus, np_minus)
            assert isinstance(rt_minus, type(np_minus))

    @given(examples=generate_sample_test_integers(num_bits=32, signed=True))
    def test_int32(self, examples):
        for i in examples:
            rt_err = None
            np_err = None
            try:
                rt_int32 = rt.int32(i)
            except BaseException as e:
                rt_err = str(e)
            try:
                np_int32 = np.int32(i)
            except BaseException as e:
                np_err = str(e)
            if rt_err or np_err:
                assert rt_err == np_err
            else:
                assert_equal_(rt_int32, np_int32)
                assert isinstance(rt_int32, type(np_int32))

                rt_plus = rt_int32 + 1
                np_plus = np_int32 + 1
                assert_equal_(rt_plus, np_plus)
                assert isinstance(rt_plus, type(np_plus))

                rt_minus = rt_int32 - 1
                np_minus = np_int32 - 1
                assert_equal_(rt_minus, np_minus)
                assert isinstance(rt_minus, type(np_minus))

    @given(examples=generate_sample_test_integers(num_bits=64, signed=True))
    def test_int64(self, examples):
        for i in examples:
            rt_err = None
            np_err = None
            try:
                rt_int64 = rt.int64(i)
            except BaseException as e:
                rt_err = str(e)
            try:
                np_int64 = np.int64(i)
            except BaseException as e:
                np_err = str(e)

            if rt_err or np_err:
                assert rt_err == np_err
            else:
                assert_equal_(rt_int64, np_int64)
                assert isinstance(rt_int64, type(np_int64))

                rt_plus = rt_int64 + 1
                np_plus = np_int64 + 1
                assert_equal_(rt_plus, np_plus)
                assert isinstance(rt_plus, type(np_plus))

                rt_minus = rt_int64 - 1
                np_minus = np_int64 - 1
                assert_equal_(rt_minus, np_minus)
                assert isinstance(rt_minus, type(np_minus))

    @given(arg_dict=interpolation_data())
    def test_interp(self, arg_dict):
        rt_err = np_err = rt_interp = np_interp = None

        try:
            rt_interp = rt.interp(
                arg_dict["x"], arg_dict["xp"], arg_dict["fp"]
            )  # According to the documentation, Riptable
            # does not implement kwargs left, right,
            # and period
        except Exception as e:
            rt_err = e

        try:
            np_interp = np.interp(arg_dict["x"], arg_dict["xp"], arg_dict["fp"])
        except Exception as e:
            np_err = e

        if rt_err or np_err:
            assert rt_err == np_err
        else:
            assert_equal_(rt_interp.dtype.name, np_interp.dtype.name)
            assert_allclose(np.array(rt_interp), np_interp)
            assert isinstance(rt_interp, FastArray)

    @given(arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes(endianness="=")))
    def test_isfinite(self, arr):
        rt_isfinite = rt.isfinite(arr)
        np_isfinite = np.isfinite(arr)
        assert_array_equal_(rt_isfinite, np_isfinite)
        assert isinstance(rt_isfinite, FastArray)

    @given(arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes(endianness="=")))
    def test_isinf(self, arr):
        rt_isinf = rt.isinf(arr)
        np_isinf = np.isinf(arr)
        assert_array_equal_(rt_isinf, np_isinf)
        assert isinstance(rt_isinf, FastArray)

    @given(arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes(endianness="=")))
    def test_isnan(self, arr):
        rt_isnan = rt.isnan(arr)
        np_isnan = np.isnan(arr)
        assert_array_equal_(rt_isnan, np_isnan)
        assert isinstance(rt_isnan, FastArray)

    @pytest.mark.xfail(reason="RIP-339: lexsort broken for ndarray and FastArray input types")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Performance issues when running on TeamCity")
    @given(data())
    def test_lexsort_basic(self, data):
        shape, dtype = (
            shared(ndarray_shape_strategy(max_rank=1)),
            shared(ints_floats_datetimes_and_timedeltas()),
        )
        a = data.draw(arrays(shape=shape, dtype=dtype))
        b = data.draw(arrays(shape=shape, dtype=dtype))

        assert_array_equal_(np.lexsort((b, a)), rt.lexsort((list(b), list(a))))

    @pytest.mark.xfail(reason="RIP-339: lexsort broken for ndarray and FastArray input types")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Performance issues when running on TeamCity")
    @given(data())
    def test_lexsort_basic_fail(self, data):
        shape, dtype = (
            shared(ndarray_shape_strategy(max_rank=1)),
            shared(ints_or_floats_dtypes()),
        )
        a = data.draw(arrays(shape=shape, dtype=dtype))
        b = data.draw(arrays(shape=shape, dtype=dtype))

        rt_lexsort_args = [(FA(b), FA(a)), (FA(b), a), (b, FA(a)), (b, a)]
        for rt_arg in rt_lexsort_args:
            assert_array_equal_(np.lexsort((b, a)), rt.lexsort(rt_arg))

    @pytest.mark.xfail(reason="RIP-339: lexsort broken for ndarray and FastArray input types")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Performance issues when running on TeamCity")
    @given(data())
    def test_lexsort_mixed(self, data):
        shape, dtype = (
            shared(ndarray_shape_strategy(max_rank=1)),
            ints_floats_datetimes_and_timedeltas(),
        )
        a = data.draw(arrays(shape=shape, dtype=dtype))
        b = data.draw(arrays(shape=shape, dtype=dtype))

        assert_array_equal_(np.lexsort((b, a)), rt.lexsort((list(b), list(a))))

    @pytest.mark.xfail(reason="RIP-339: lexsort broken for ndarray and FastArray input types")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Performance issues when running on TeamCity")
    @given(data())
    def test_lexsort_mixed_fail(self, data):
        shape, dtype = (
            shared(ndarray_shape_strategy(max_rank=1)),
            ints_floats_datetimes_and_timedeltas(),
        )
        a = data.draw(arrays(shape=shape, dtype=dtype))
        b = data.draw(arrays(shape=shape, dtype=dtype))

        rt_lexsort_args = [(FA(b), FA(a)), (FA(b), a), (b, FA(a)), (b, a)]
        for rt_arg in rt_lexsort_args:
            assert_array_equal_(np.lexsort((b, a)), rt.lexsort(rt_arg))

    @pytest.mark.xfail(reason="RIP-339: lexsort broken for ndarray and FastArray input types")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Performance issues when running on TeamCity")
    @given(data())
    def test_lexsort_many_columns(self, data):
        k = data.draw(integers(min_value=1, max_value=_MAX_SHAPE_SIZE // 10))
        shape, dtype = (
            shared(ndarray_shape_strategy(max_rank=1)),
            shared(integer_dtypes(endianness="=")),
        )

        keys: List[List[int]] = list()
        for _ in range(k):
            keys.append(list(data.draw(arrays(shape=shape, dtype=dtype))))

        assert_array_equal_(np.lexsort(keys), rt.lexsort(keys))

    @pytest.mark.xfail(reason="RIP-339: lexsort broken for ndarray and FastArray input types")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Performance issues when running on TeamCity")
    @given(data())
    def test_lexsort_many_columns_fail(self, data):
        k = data.draw(integers(min_value=1, max_value=_MAX_SHAPE_SIZE // 10))
        shape, dtype = (
            shared(ndarray_shape_strategy(max_rank=1)),
            shared(integer_dtypes(endianness="=")),
        )

        keys: List[FastArray] = list()
        for _ in range(k):
            keys.append(FA(data.draw(arrays(shape=shape, dtype=dtype))))

        assert_array_equal_(np.lexsort(keys), rt.lexsort(keys))

    @given(arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes(endianness="=")))
    def test_log_array(self, arr):
        rt_log = rt.log(arr)
        np_log = np.log(arr)
        assert_allclose(np.array(rt_log), np_log, rtol=1e-6)
        assert isinstance(rt_log, FastArray)

    @given(scalar=floats())
    def test_log_scalar(self, scalar):
        rt_log = rt.log(scalar)
        np_log = np.log(scalar)
        assert_allclose(rt_log, np_log, rtol=1e-6)

    @given(arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes(endianness="=")))
    def test_log10_array(self, arr):
        rt_log10 = rt.log10(arr)
        np_log10 = np.log10(arr)
        assert_allclose(np.array(rt_log10), np_log10, rtol=1e-6)
        assert isinstance(rt_log10, FastArray)

    @given(scalar=floats())
    def test_log10_scalar(self, scalar):
        rt_log10 = rt.log10(scalar)
        np_log10 = np.log10(scalar)
        assert_allclose(rt_log10, np_log10, rtol=1e-6)

    @pytest.mark.xfail(reason="This test exposes a known bug around NaN-handling in 'max' that needs to be fixed.")
    @given(
        arr_axis_tuple=generate_array_and_axis(
            shape=ndarray_shape_strategy(), dtype=ints_floats_datetimes_and_timedeltas()
        )
    )
    @given(mats=same_structure_ndarrays())
    def test_maximum(self, mats):
        mat1, mat2 = mats

        np_output = np.maximum(mat1, mat2)
        rt_output = rt.maximum(mat1, mat2)

        assert_array_equal_(np.array(rt_output), np_output)

        # TODO: Enable this assertion once we've fixed riptable so rt.maximum returns a FastArray.
        # assert isinstance(rt_output, FastArray)

    @given(
        arr_axis_tuple=generate_array_and_axis(
            shape=ndarray_shape_strategy(), dtype=ints_floats_datetimes_and_timedeltas()
        )
    )
    def test_min(self, arr_axis_tuple):
        arr, axis = arr_axis_tuple

        rt_output = rt.min(arr, axis=axis)
        np_output = np.min(arr, axis=axis)

        assert_array_equal_(np.array(rt_output), np_output)

        # TODO: Enable this assertion once we've fixed riptable so rt.maximum returns a FastArray.
        # if axis: #if axis is None, it will return a scalar
        #     assert isinstance(rt_output, FastArray)

    @pytest.mark.parametrize(
        "func",
        [
            # TODO add params: nanmean, nansum, nanstd, nanvar, nanargmin, nanargmax
            # TODO add other missing array methods, then dynamically generate these functions from introspection of numpy and riptable
            # List of ndarray array methods: https://docs.scipy.org/doc/numpy/reference/arrays.ndarray.html#array-methods
            "abs",
            "all",
            "any",
            "argmax",
            "argmin",
            "argsort",
            # TODO add cumprod
            "cumsum",
            "mean",
            "median",
            "nanmedian",
            "percentile",
            "std",
            "var",
            pytest.param(
                "sum",
                marks=[
                    pytest.mark.xfail(
                        reason="Need to investigate intermittent failures possibly due to roundoff errors."
                    )
                ],
            ),
            pytest.param(
                "nanmax",
                marks=[
                    pytest.mark.xfail(
                        reason="RIP-417: This test exposes a known bug around NaN-handling in 'nanmax' that needs to be fixed."
                    )
                ],
            ),
            pytest.param(
                "nanmin",
                marks=[
                    pytest.mark.xfail(
                        reason="RIP-417: This test exposes a known bug around NaN-handling in 'nanmin' that needs to be fixed."
                    )
                ],
            ),
            pytest.param(
                "max",
                marks=[
                    pytest.mark.xfail(
                        reason="Conflict in comparing FastArray with invalids that don't exist as ndarray invalids. Need to handle invalid assertion checks."
                    )
                ],
            ),
            pytest.param(
                "nanpercentile",
                marks=[
                    pytest.mark.xfail(
                        reason="Investigate frequent TeamCity hypothesis DeadlineExceeded exceptions that never occur on local runs"
                    )
                ],
            ),
        ],
    )
    @pytest.mark.parametrize(
        "include_invalid",
        [False],  # TODO enable including invalids and use masked arrays to check valid and possibly invalid values
    )
    def test_array_methods(self, func, include_invalid):
        # TODO rework so the generated array and axis can be cached and reused, otherwise most of the time spent for this test is generating arrays and an axis
        # Attempted to use the hypothesis cacheable decorator, but it doesn't cache the strategy for reuse
        # Investigate reworking generate_array_and_axis by using the defines_strategy_with_reusable_values decorator
        @given(
            arr_axis_tuple=generate_array_and_axis(
                shape=ndarray_shape_strategy(),
                dtype=ints_or_floats_dtypes(),
                include_invalid=include_invalid,
            )
        )
        def inner(arr_axis_tuple):
            arr, axis = arr_axis_tuple

            # TODO enable this assertion for guarded funcs once we've fixed riptable so these return a fastarray.
            def _check_fastarray(inst):
                if axis:  # need an axis to check if output is a FastArray, otherwise output is a scalar
                    assert isinstance(inst, rt.FastArray)

            args = (func, arr)
            kwargs = {}
            if func not in ["abs"]:
                kwargs["axis"] = axis
            # set up keyword arguments needed for some of the rt_numpy functions
            if "std" in func or "var" in func:
                kwargs["ddof"] = 1
            if "percentile" in func:
                kwargs[QUANTILE_METHOD_NP_KW] = "linear"
                kwargs["q"] = 50
            if "argsort" == func:
                kwargs["kind"] = "stable"

            # Test #1:  Check calling .func() as a module function in numpy vs. riptable.
            rt_output = operator.methodcaller(*args, **kwargs)(rt)
            np_output = operator.methodcaller(*args, **kwargs)(np)
            assert_array_equal_(rt_output, np_output)

            # Test #2:  Check calling .func() as a module function in numpy vs. riptable.
            #           This time, explicitly wrap the ndarray generated by hypothesis in a FastArray
            #           so we're sure to get the FastArray implementation.
            f_arr = rt.FA(arr)
            args = (func, f_arr)
            rt_output = operator.methodcaller(*args, **kwargs)(rt)
            np_output = operator.methodcaller(*args, **kwargs)(np)
            assert_array_equal_(rt_output, np_output)
            # _check_fastarray(rt_output)  # rt_output should be a FastArray, but is a ndarray

            # Test #3:  Call the .func() method on both the FastArray and the ndarray.
            # Either FastArray or ndarray does not support the guarded member names.
            common_members, _ = _NDARRAY_FASTARRAY_COMMON_AND_DIFF
            if func in common_members:
                rt_output = operator.methodcaller(func, **kwargs)(f_arr)
                np_output = operator.methodcaller(func, **kwargs)(arr)
                assert_array_equal_(rt_output, np_output)
                # _check_fastarray(rt_output)  # rt_output should be a FastArray, but is a ndarray

        inner()

    @given(arrs=sampled_from([[9, 8, 7], [[10, 1], [11, 2], [13, 3]]]), inds=just([False, True, False]))
    def test_array_boolean_indexing(self, arrs, inds):
        np_array = np.asfortranarray(arrs)
        rt_array = rt.FA(np_array)

        np_bool_indices = inds
        rt_bool_indices = rt.FA(np_bool_indices)

        nr = np_array[np_bool_indices]
        fr = rt_array[rt_bool_indices]

        assert_array_equal_(np.array(fr), nr)

    @given(mats=same_structure_ndarrays())
    def test_minimum(self, mats):
        mat1, mat2 = mats

        np_output = np.minimum(mat1, mat2)
        rt_output = rt.minimum(mat1, mat2)

        assert_array_equal_(np.array(rt_output), np_output)

        # TODO: Enable this assertion once we've fixed riptable so rt.minimum returns a FastArray.
        # assert isinstance(rt_output, FastArray)

    @pytest.mark.xfail(
        reason="SOQTEST-6497 Riptable.nan_to_num() does not return a FastArray",
        raises=AssertionError,
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(array=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes()))
    def test_nan_to_num_array(self, array):
        rt_nan_to_num = rt.nan_to_num(array)
        np_nan_to_num = np.nan_to_num(array)

        assert_allclose(rt_nan_to_num, np_nan_to_num)
        assert isinstance(rt_nan_to_num, FastArray)

    @given(scalar=floats())
    def test_nan_to_num_scalar(self, scalar):
        rt_nan_to_num = rt.nan_to_num(scalar)
        np_nan_to_num = np.nan_to_num(scalar)

        assert_allclose(rt_nan_to_num, np_nan_to_num)

    @given(
        arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes()),
        q=floats(min_value=0, max_value=100),
    )
    def test_nanpercentile_single(self, arr, q):
        interpolations = ["linear", "lower", "higher", "midpoint", "nearest"]
        for interpolation in interpolations:
            # todo - rt_nanpercentile should come from riptable
            kwargs = {QUANTILE_METHOD_NP_KW: interpolation}
            rt_nanpercentile = np.nanpercentile(arr, q=q, **kwargs)
            np_nanpercentile = np.nanpercentile(arr, q=q, **kwargs)
            assert_allclose(rt_nanpercentile, np_nanpercentile)

    # @given(arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes()), q=floats(min_value=0, max_value=100))
    # def test_nanpercentile_single(self, arr, q):
    #     interpolations = ['linear', 'lower', 'higher', 'midpoint', 'nearest']
    #     for interpolation in interpolations:
    #         # todo - rt_nanpercentile should come from riptable
    #         rt_nanpercentile = np.nanpercentile(arr, q=q, interpolation=interpolation)
    #         np_nanpercentile = np.nanpercentile(arr, q=q, interpolation=interpolation)
    #         assert_allclose(rt_nanpercentile, np_nanpercentile)

    @pytest.mark.xfail(
        reason="SOQTEST-6497 Riptable.nanpercentile() does not return a FastArray",
        raises=AssertionError,
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes()),
        q=lists(floats(min_value=0, max_value=100), min_size=10, max_size=50, unique=True),
    )
    def test_nanpercentile_array(self, arr, q):
        interpolations = ["linear", "lower", "higher", "midpoint", "nearest"]
        for interpolation in interpolations:
            kwargs = {QUANTILE_METHOD_NP_KW: interpolation}
            # todo - rt_nanpercentile should come from riptable
            rt_output = np.nanpercentile(arr, q=q, **kwargs)
            np_output = np.nanpercentile(arr, q=q, **kwargs)

            assert_allclose(rt_output, np_output)
            if len(arr.shape) > 1:
                assert isinstance(rt_output, FastArray)

    @pytest.mark.xfail(
        reason="SOQTEST-6637 Riptable does not convert the array to a FastArray, so it does not guarantee nanstd will be an available attribute\n"
        "SOQTEST-6497 Riptable.nanstd() does not return a FastArray",
        raises=(AttributeError, AssertionError),
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(arr_axis_and_ddof=generate_array_axis_and_ddof())
    def test_nanstd(self, arr_axis_and_ddof):
        """
        This function has a different default behavior from numpy. Riptable, like pandas, has a default ddof of 1, while
        numpy uses 0.
        """
        arr = arr_axis_and_ddof["arr"]
        axis = arr_axis_and_ddof["axis"]
        ddof = arr_axis_and_ddof["ddof"]

        rt_output = rt.nanstd(arr, axis=axis, ddof=ddof)
        np_output = np.nanstd(arr, axis=axis, ddof=ddof)

        assert_allclose(rt_output, np_output)
        if axis and len(arr.shape) > 1:
            assert isinstance(rt_output, FastArray)

    @pytest.mark.xfail(
        reason="SOQTEST-6637 Riptable does not convert the array to a FastArray, so it does not guarantee nansum will be an available attribute\n"
        "nansum does not implement the axis argument either",
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        array_and_axis=generate_array_and_axis(shape=ndarray_shape_strategy(max_rank=1), dtype=ints_or_floats_dtypes())
    )
    def test_nansum(self, array_and_axis):
        """
        riptable.nansum() does not yet implement the axis argument, since riptable is mostly focused on 1D FastArrays.
        test_nansum_1D only selects arrays that are rank 0 or 1. Riptable's nansum, like nanmean, does not cast the array
        passed in to a FastArray--when a FastArray is not passed in an Attribute exception occurs because the function
        tries calling array.nansum().
        """
        arr, axis = array_and_axis
        arr = FastArray(arr)

        rt_output = rt.nansum(arr, axis=axis)
        np_output = np.nansum(arr, axis=axis)

        assert_allclose(rt_output, np_output)
        if axis:
            assert isinstance(rt_output, FastArray)

    # The floats should have the same byte-order as the machine at the moment (SOQTEST-6478) -> endianness="="
    # and riptable only officially supports float32 and float64 floating-point datatypes, so using other ones occasionally
    # leads to errors.
    # Split up ints and floats to ensure the same datatypes for widening. (int->int64, float->float64)
    @pytest.mark.skip(
        reason="When the array has very large numbers and numbers closed to zero, numpy and riptable can produce different results."
    )
    @given(
        arr=arrays(
            shape=ndarray_shape_strategy(max_rank=1),
            dtype=floating_dtypes(endianness="=", sizes=(32, 64)),
        )
    )
    def test_nansum_1D_float(self, arr):
        arr = FastArray(arr)
        # TODO: numpy is not widening the output array as expected to 64 bits all the time, so it is being forced here.
        rt_output = rt.nansum(arr, axis=None, dtype=np.float64)
        np_output = np.nansum(arr, axis=None, dtype=np.float64)
        assert_allclose(rt_output, np_output)

    # The floats should have the same byte-order as the machine at the moment (SOQTEST-6478) -> endianness="="
    # Split up ints and floats to ensure the same datatypes for widening. (int->int64, float->float64)
    @pytest.mark.xfail(reason="SOQTEST-6670 Nansum does not follow the dtype argument and returns a float64")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        arr=arrays(
            shape=ndarray_shape_strategy(max_rank=1),
            dtype=integer_dtypes(endianness="="),
        )
    )
    def test_nansum_1D_int(self, arr):
        arr = FastArray(arr)
        # TODO: numpy is not widening the output array as expected to 64 bits all the time, so it is being forced here.
        rt_output = rt.nansum(arr, axis=None, dtype=np.int64)
        np_output = np.nansum(arr, axis=None, dtype=np.int64)
        assert rt_output.dtype.type == np.int64
        assert_allclose(rt_output, np_output)

    @pytest.mark.xfail(
        reason="SOQTEST-6637 Riptable does not convert the array to a FastArray, so it does not guarantee nanvar will be an available attribute\n"
        "SOQTEST-6497 Riptable.nanvar() does not return a FastArray",
        raises=(AttributeError, AssertionError),
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(arr_axis_and_ddof=generate_array_axis_and_ddof())
    def test_nanvar(self, arr_axis_and_ddof):
        arr = arr_axis_and_ddof["arr"]
        axis = arr_axis_and_ddof["axis"]
        ddof = arr_axis_and_ddof["ddof"]

        rt_output = rt.nanvar(arr, axis=axis, ddof=ddof)
        np_output = np.nanvar(arr, axis=axis, ddof=ddof)

        assert_allclose(rt_output, np_output)
        if axis and len(arr.shape) > 1:
            assert isinstance(rt_output, FastArray)

    @given(
        shape=ndarray_shape_strategy(),
        dtype=ints_or_floats_dtypes(),
        order=sampled_from(("F", "C")),
    )
    def test_ones(self, shape, dtype, order):
        rt_ones = rt.ones(shape, dtype, order)
        np_ones = np.ones(shape=shape, dtype=dtype, order=order)
        assert isinstance(rt_ones, FastArray)

        assert_equal_(shape, rt_ones.shape)
        assert_equal_(dtype.type, rt_ones.dtype.type)
        assert_array_equal_(rt_ones, np_ones)

        # 1-D arrays always use Column-order. Otherwise, use the order specified
        if len(rt_ones.shape) > 1 and min(rt_ones.shape) > 1:
            assert np.isfortran(rt_ones) == (order == "F")
        else:
            assert not np.isfortran(rt_ones)

    @pytest.mark.xfail(reason="SOQTEST-6495 kwargs not implemented in riptable.ones")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        shape=ndarray_shape_strategy(),
        dtype=ints_or_floats_dtypes(),
        order=sampled_from(("F", "C")),
    )
    def test_ones_kwargs(self, shape, dtype, order):
        rt_ones = rt.ones(shape=shape, dtype=dtype, order=order)
        np_ones = np.ones(shape=shape, dtype=dtype, order=order)
        assert isinstance(rt_ones, FastArray)

        assert_equal_(shape, rt_ones.shape)
        assert_equal_(dtype.type, rt_ones.dtype.type)
        assert_array_equal_(rt_ones, np_ones)

        # 1-D arrays always use Column-order. Otherwise, use the order specified
        if len(rt_ones.shape) > 1 and min(rt_ones.shape) > 1:
            assert np.isfortran(rt_ones) == (order == "F")
        else:
            assert not np.isfortran(rt_ones)

    @pytest.mark.xfail(reason="SOQTEST-6563 riptable does not implement subok")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        arr=arrays(shape=ndarray_shape_strategy(), dtype=ints_floats_datetimes_and_timedeltas()),
        dtype=ints_floats_or_complex_dtypes(),
        order=sampled_from(("C", "F", "K", "A")),
        subok=booleans(),
    )
    def test_ones_like(self, arr, dtype, order, subok):
        rt_ones_like = rt.ones_like(arr, dtype=dtype, order=order, subok=subok)
        np_ones_like = np.ones_like(arr, dtype=dtype, order=order, subok=subok)

        assert_equal_(dtype, rt_ones_like.dtype)
        assert_equal_(rt_ones_like.shape, arr.shape)
        assert_array_equal_(rt_ones_like, np_ones_like)

        # 1-D arrays always use Column-order. Otherwise, use the order specified
        if (
            len(rt_ones_like.shape) > 1
            and min(rt_ones_like.shape) > 1
            and (order == "F" or ((order == "A" or order == "K") and np.isfortran(arr)))
        ):
            assert np.isfortran(rt_ones_like)
        else:
            assert not np.isfortran(rt_ones_like)

        if subok:
            assert isinstance(rt_ones_like, FastArray)
        else:
            assert isinstance(rt_ones_like, np.ndarray)
            # FastArray is a subclass of np.ndarray, so also ensure it is not a FastArray
            assert not isinstance(rt_ones_like, FastArray)

    @given(
        arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes()),
        q=floats(min_value=0, max_value=100),
    )
    def test_percentile_single(self, arr, q):
        interpolations = ["linear", "lower", "higher", "midpoint", "nearest"]
        for interpolation in interpolations:
            kwargs = {QUANTILE_METHOD_NP_KW: interpolation}
            # todo - rt_percentile should come from riptable
            rt_output = np.percentile(arr, q=q, axis=None, **kwargs)
            np_output = np.percentile(arr, q=q, axis=None, **kwargs)
            assert_allclose(rt_output, np_output)

    @pytest.mark.xfail(
        reason="SOQTEST-6497 Riptable.percentile() does not return a FastArray",
        raises=AssertionError,
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        array_and_axis=generate_array_and_axis(shape=ndarray_shape_strategy(), dtype=floating_dtypes()),
        q=lists(floats(min_value=0, max_value=100), min_size=10, max_size=50, unique=True),
    )
    def test_percentile_array(self, array_and_axis, q):
        arr, axis = array_and_axis
        interpolations = ["linear", "lower", "higher", "midpoint", "nearest"]
        for interpolation in interpolations:
            kwargs = {QUANTILE_METHOD_NP_KW: interpolation}
            rt_output = rt.percentile(arr, q=q, axis=axis, **kwargs)
            np_output = np.percentile(arr, q=q, axis=axis, **kwargs)
            assert_allclose(rt_output, np_output)
            if axis and len(arr.shape) > 1:
                assert isinstance(rt_output, FastArray)

    @pytest.mark.xfail(
        reason="SOQTEST-6497 Riptable.putmask does not cast input to a FastArray",
        raises=AssertionError,
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        array_and_mask=generate_array_and_where(shape=ndarray_shape_strategy(), dtype=floating_dtypes(endianness="=")),
        rep_val=floats(),
    )
    def test_putmask(self, array_and_mask, rep_val):
        arr, mask = array_and_mask
        rt_err = None
        rt_putmask = arr.copy()
        np_err = None
        np_putmask = arr.copy()
        try:
            rt.putmask(rt_putmask, mask, rep_val)
        except OverflowError:
            rt_err = OverflowError

        try:
            np.putmask(np_putmask, mask, rep_val)
        except OverflowError:
            np_err = OverflowError

        if rt_err or np_err:
            assert rt_err == np_err
        else:
            assert_allclose(rt_putmask, np_putmask)
            assert isinstance(rt_putmask, FastArray)

    @pytest.mark.xfail(
        reason="RIP-341: quantile's _get_score raises an index error when getting the view from _val and using it to access the ndarray."
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(arr=arrays(shape=one_darray_shape_strategy(), dtype=floating_dtypes(sizes=(32, 64))))
    def test_quantile(self, arr):
        # For a one to one equivalence of supported interpolation methods between numpy and riptable:
        # 1) np.quantile would need to support riptable's fraction method
        # 2) rt.quantile would need to support numpy's linear, midpoint, and nearest methods
        # Numpy quantile reference https://docs.scipy.org/doc/numpy/reference/generated/numpy.quantile.html
        # Tracked in Jira RIP-344
        interpolations = ["lower", "higher"]
        for interpolation in interpolations:
            q = random.random()
            rt_quantile, _, _ = rt.quantile(arr, q, interpolation)
            np_quantile = np.quantile(arr, q, interpolation=interpolation)
            assert_allclose(rt_quantile, np_quantile, err_msg=f"to reproduce use quantile {q}")

    @given(
        arr=arrays(shape=ndarray_shape_strategy(), dtype=ints_or_floats_dtypes()),
        rep=integers(min_value=0, max_value=10),
        axis=sampled_from((None, 0, 1)),
    )
    def test_repeat_integer_repeats(self, arr, rep, axis):
        if axis == 1 and len(arr.shape) == 1:
            axis = 0

        rt_repeat = rt.repeat(arr, repeats=rep, axis=axis)
        np_repeat = np.repeat(arr, repeats=rep, axis=axis)
        assert_allclose(np.array(rt_repeat), np_repeat)

    @given(
        arr_axis_and_rep=generate_array_axis_and_repeats_array(
            dtype=ints_or_floats_dtypes(), shape=ndarray_shape_strategy()
        )
    )
    def test_repeat_array_repeats(self, arr_axis_and_rep):
        arr, axis, rep = arr_axis_and_rep

        rt_repeat = rt.repeat(arr, repeats=rep, axis=axis)
        np_repeat = np.repeat(arr, repeats=rep, axis=axis)

        assert_allclose(np.array(rt_repeat), np_repeat)
        assert isinstance(rt_repeat, FastArray)

    @given(
        arr_and_dim=generate_reshape_array_and_shape_strategy(
            shape=ndarray_shape_strategy(), dtype=ints_or_floats_dtypes()
        )
    )
    def test_reshape(self, arr_and_dim):
        arr, dim = arr_and_dim
        orders = ["C", "F", "A"]
        for order in orders:
            rt_reshape = rt.reshape(arr, dim, order=order)
            np_reshape = np.reshape(arr, dim, order=order)
            assert_allclose(np.array(rt_reshape), np_reshape)
            assert np.isfortran(rt_reshape) == np.isfortran(np_reshape)
            assert isinstance(rt_reshape, FastArray)

    @pytest.mark.xfail(
        reason="SOQTEST-6530 riptable.round() does not return FastArray and decimals arg not yet implemented"
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes(sizes=(32, 64))))
    def test_round_array(self, arr):
        # TODO: Use range of decimals once SOQTEST-6530 is addressed
        decimals = [0]
        # decimals = range(-5, 100)

        for decimal in decimals:
            # test array
            rt_output = rt.round(arr, decimals=decimal)
            np_output = np.round(arr, decimals=decimal)

            assert_array_equal_(np.array(rt_output), np_output)
            assert isinstance(rt_output, FastArray)

    @pytest.mark.xfail(
        reason="SOQTEST-6530 should fail on +/-inf, and [-0.5,-0], since riptable.round calls builtin round. Decimals arg also not yet implemented"
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(scalar=floating_scalar())
    def test_round_scalar(self, scalar):
        # TODO: Use range of decimals once SOQTEST-6530 is addressed
        decimals = [0]
        # decimals = range(-5, 100)
        for decimal in decimals:
            # test scalars
            rt_round_scalar = rt.round(scalar, decimals=decimal)
            np_round_scalar = np.round(scalar, decimals=decimal)
            assert_equal_(
                rt_round_scalar,
                np_round_scalar,
                err_msg=f"Rounding error on {type(scalar)} {scalar}",
            )

    @pytest.mark.xfail(reason="RIP-345 - discrepancy between inserts position points")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(arrays(shape=one_darray_shape_strategy(), dtype=integer_dtypes()))
    def test_searchsorted(self, arr):
        sides = {"left", "right"}

        arr.sort()
        v = arr[arr % 2 == 0]

        for side in sides:
            rt_indicies = rt.searchsorted(arr, v, side)
            np_indicies = np.searchsorted(arr, v, side)
            assert_array_equal_(
                rt_indicies,
                np_indicies,
                err_msg=f"using array {arr}\nvalues to insert {v}\nside {side}",
            )

    @pytest.mark.xfail(
        reason="RIP-350 - see Python/core/riptable/tests/test_base_function.TestStd.test_std for a materialized counterexample"
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        arr=arrays(
            shape=one_darray_shape_strategy(max_shape_size=10),
            # do not consider overflow; numpy and riptable differ in how they behave on overflow
            elements=integers(-int(_MAX_INT / 10), int(_MAX_INT / 10)),
            dtype=integer_dtypes(endianness="=", sizes=(64,)),
        )
    )
    def test_std(self, arr):
        np_std = np.std(arr, ddof=1)
        rt_std = rt.std(rt.FastArray(arr))
        assert_equal_(rt_std, np_std, decimal=6)

    @given(
        arr=arrays(
            shape=one_darray_shape_strategy(max_shape_size=10),
            # do not consider overflow; numpy and riptable differ in how they behave on overflow
            elements=integers(-int(_MAX_INT / 10), int(_MAX_INT / 10)),
            dtype=integer_dtypes(endianness="=", sizes=(64,)),
        )
    )
    def test_sum_int(self, arr):
        np_sum = np.sum(arr)
        rt_sum = rt.sum(rt.FastArray(arr))
        assert_equal_(int(rt_sum), int(np_sum))

    @pytest.mark.xfail(
        reason="RIP-351 - see Python/core/riptable/tests/test_base_function.TestSum.test_sum_float for a materialized counterexample"
    )
    @given(
        arr=arrays(
            shape=one_darray_shape_strategy(max_shape_size=10),
            # do not consider overflow; numpy and riptable differ in how they behave on overflow
            elements=floats(-int(_MAX_FLOAT / 10), int(_MAX_FLOAT / 10)),
            dtype=floating_dtypes(endianness="=", sizes=(64,)),
        )
    )
    def test_sum_float(self, arr):
        np_sum = np.sum(arr)
        rt_sum = rt.sum(rt.FastArray(arr))
        assert_equal_(rt_sum, np_sum, decimal=6)

    @pytest.mark.skip(reason="Resolve DeadlineExceeded errors on workflow runs")
    @given(
        arr=arrays(
            shape=ndarray_shape_strategy(max_shape_size=5),
            dtype=floating_dtypes(endianness="=", sizes=(32, 64)),
        ),
        shape=ndarray_shape_strategy(),
    )
    def test_tile(self, arr, shape):
        np_result = np.tile(arr, shape)
        rt_result = rt.tile(arr, shape)
        assert_array_equal_(rt_result, np_result)

    @pytest.mark.xfail(reason="RIP-358 discrepency between rt.trunc and np.trunc")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        data=data(),
        arr=arrays(shape=ndarray_shape_strategy(), dtype=floating_dtypes(sizes=(32, 64))),
    )
    def test_trunc(self, data, arr):
        # TODO: Modify this to use the 'generate_array_and_where' strategy instead?
        where_arr = data.draw(arrays(shape=arr.shape, dtype=boolean_dtypes()))

        # Test #1: Returns new array containing results; no optional parameters provided.
        rt_output = rt.trunc(arr)
        np_output = np.trunc(arr)
        assert isinstance(rt_output, FastArray)
        assert_array_equal_(np.array(rt_output), np_output)

        # Test #2: Returns new array containing results; 'where' bitmask is provided.
        rt_output = rt.trunc(arr, where=where_arr)
        np_output = np.trunc(arr, where=where_arr)
        assert isinstance(rt_output, FastArray)
        assert_array_equal_(np.array(rt_output), np_output)

        # Test #3: Results written to array specified in 'out' parameter; no other optional params.
        rt_inplace = np.zeros_like(arr)
        np_inplace = np.zeros_like(arr)
        rt_output = rt.trunc(arr, out=rt_inplace)
        np_output = np.trunc(arr, out=np_inplace)
        assert_array_equal_(np.array(rt_inplace), np_inplace)
        # TODO: Add assertions for rt_output and np_output -- what are they expected to return when the 'out' parameter is specified?
        # assert isinstance(rt_output, FastArray, msg="riptable.trunc() did not return a FastArray")

        # Test #4: Results written to array specified in 'out' parameter; 'where' bitmask is provided.
        rt_inplace = np.zeros_like(arr)
        np_inplace = np.zeros_like(arr)
        rt_output = rt.trunc(arr, where=where_arr, out=rt_inplace)
        np_output = np.trunc(arr, where=where_arr, out=np_inplace)
        assert_array_equal_(np.array(rt_inplace), np_inplace)
        # TODO: Add assertions for rt_output and np_output -- what are they expected to return when the 'out' parameter is specified?
        # assert isinstance(rt_output, FastArray, msg="riptable.trunc() did not return a FastArray")

    @given(scalar=floating_scalar())
    def test_trunc_scalar(self, scalar):
        # test scalars
        rt_trunc_scalar = rt.trunc(scalar)
        np_trunc_scalar = np.trunc(scalar)
        assert_equal_(rt_trunc_scalar, np_trunc_scalar)

    @given(examples=generate_sample_test_integers(num_bits=8, signed=False))
    def test_uint8(self, examples):
        for i in examples:
            rt_uint8 = rt.uint8(i)
            np_uint8 = np.uint8(i)
            assert_equal_(rt_uint8, np_uint8)
            assert isinstance(rt_uint8, type(np_uint8))

            rt_plus = rt_uint8 + 1
            np_plus = np_uint8 + 1
            assert_equal_(rt_plus, np_plus)
            assert isinstance(rt_plus, type(np_plus))

            rt_minus = rt_uint8 - 1
            np_minus = np_uint8 - 1
            assert_equal_(rt_minus, np_minus)
            assert isinstance(rt_minus, type(np_minus))

    @given(examples=generate_sample_test_integers(num_bits=16, signed=False))
    def test_uint16(self, examples):
        for i in examples:
            rt_uint16 = rt.uint16(i)
            np_uint16 = np.uint16(i)
            assert_equal_(rt_uint16, np_uint16)
            assert isinstance(rt_uint16, type(np_uint16))

            rt_plus = rt_uint16 + 1
            np_plus = np_uint16 + 1
            assert_equal_(rt_plus, np_plus)
            assert isinstance(rt_plus, type(np_plus))

            rt_minus = rt_uint16 - 1
            np_minus = np_uint16 - 1
            assert_equal_(rt_minus, np_minus)
            assert isinstance(rt_minus, type(np_minus))

    @given(examples=generate_sample_test_integers(num_bits=32, signed=False))
    def test_uint32(self, examples):
        for i in examples:
            rt_uint32 = rt.uint32(i)
            np_uint32 = np.uint32(i)
            assert_equal_(rt_uint32, np_uint32)
            assert isinstance(rt_uint32, type(np_uint32))

            rt_plus = rt_uint32 + 1
            np_plus = np_uint32 + 1
            assert_equal_(rt_plus, np_plus)
            assert isinstance(rt_plus, type(np_plus))

            rt_minus = rt_uint32 - 1
            np_minus = np_uint32 - 1
            assert_equal_(rt_minus, np_minus)
            assert isinstance(rt_minus, type(np_minus))

    @given(examples=generate_sample_test_integers(num_bits=64, signed=False))
    def test_uint64(self, examples):
        for i in examples:
            rt_uint64 = rt.uint64(i)
            np_uint64 = np.uint64(i)
            assert_equal_(rt_uint64, np_uint64)
            assert isinstance(rt_uint64, type(np_uint64))

            rt_plus = rt_uint64 + 1
            np_plus = np_uint64 + 1
            assert_equal_(rt_plus, np_plus)
            assert isinstance(rt_plus, type(np_plus))

            rt_minus = rt_uint64 - 1
            np_minus = np_uint64 - 1
            assert_equal_(rt_minus, np_minus)
            assert isinstance(rt_minus, type(np_minus))

    @given(
        arr=arrays(
            shape=one_darray_shape_strategy(),
            elements=integers(-_MAX_INT, _MAX_INT),
            dtype=integer_dtypes(endianness="=", sizes=(64,)),
        )
    )
    def test_unique(self, arr):
        rt_result = rt.unique(arr)
        np_result = np.unique(arr)
        assert_array_equal_(
            rt_result,
            np_result,
            err_msg=f"arr dtype {arr.dtype} {arr}\nrt {rt_result}\nnp {np_result}",
        )

    @hypothesis.settings(suppress_health_check=[HealthCheck.too_slow])
    @pytest.mark.xfail(
        reason="SOQTEST-6548 Riptable.vstack transposes arrays with shape (x,) and multiple (1,1)s. vstack also does not convert non-numbers to a FastArray"
    )
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(tuple_of_arrays=generate_tuples_of_arrays(all_same_width=True))
    def test_vstack(self, tuple_of_arrays):
        rt_err_type = None
        np_err_type = None
        rt_vstack = None
        np_vstack = None
        try:
            rt_vstack = rt.vstack(tuple_of_arrays)
        except ValueError:
            rt_err_type = ValueError
        try:
            np_vstack = np.vstack(tuple_of_arrays)
        except ValueError:
            np_err_type = ValueError

        if rt_err_type and np_err_type:
            assert rt_err_type == np_err_type
        else:
            assert isinstance(rt_vstack, FastArray)
            assert_array_equal_(np.array(rt_vstack), np_vstack)

    @pytest.mark.xfail(reason="See TestWhere.test_where for a materialized counterexample")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        arr=arrays(
            shape=ndarray_shape_strategy(),
            dtype=integer_dtypes(endianness="=", sizes=(32, 64)),
        )
    )
    def test_where(self, arr):
        min, mean, max = arr.min(), arr.mean(), arr.max()
        # Break this out into pytest cases; RiptableNumpyEquivalencyTests inherit from unittest
        # which does not allow for pytest.mark.parametrize.
        np_gt_min, rt_gt_min = np.where(arr > min), rt.where(arr > min)
        assert_array_equal_(
            np_gt_min,
            rt_gt_min,
            err_msg=f"array elements greater than the minimum {min}",
        )

        np_lt_mean, np_lt_mean = np.where(arr < mean), rt.where(arr < mean)
        assert_array_equal_(np_lt_mean, np_lt_mean, err_msg=f"array elements less than the mean {mean}")

        np_gt_mean, np_gt_mean = np.where(arr > mean), rt.where(arr > mean)
        assert_array_equal_(
            np_gt_mean,
            np_gt_mean,
            err_msg=f"array elements greater than the mean {mean}",
        )

        np_lt_max, np_lt_max = np.where(arr < max), rt.where(arr < max)
        assert_array_equal_(np_lt_max, np_lt_max, err_msg=f"array elements less than the max {max}")

    @given(
        shape=ndarray_shape_strategy(),
        dtype=ints_or_floats_dtypes(),
        order=sampled_from(("F", "C")),
    )
    def test_zeros(self, shape, dtype, order):
        rt_zeros = rt.zeros(shape, dtype, order)
        np_zeros = np.zeros(shape=shape, dtype=dtype, order=order)
        assert isinstance(rt_zeros, FastArray)

        assert_equal_(shape, rt_zeros.shape)
        assert_equal_(dtype.type, rt_zeros.dtype.type)
        assert_array_equal_(rt_zeros, np_zeros)

        # 1-D arrays always use Column-order. Otherwise, use the order specified
        if len(rt_zeros.shape) > 1 and min(rt_zeros.shape) > 1:
            assert np.isfortran(rt_zeros) == (order == "F")
        else:
            assert not np.isfortran(rt_zeros)

    @pytest.mark.xfail(reason="SOQTEST-6495 kwargs not implemented in riptable.zeros")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        shape=ndarray_shape_strategy(),
        dtype=ints_or_floats_dtypes(),
        order=sampled_from(("F", "C")),
    )
    def test_zeros_kwargs(self, shape, dtype, order):
        rt_zeros = rt.zeros(shape=shape, dtype=dtype, order=order)
        np_zeros = np.zeros(shape=shape, dtype=dtype, order=order)
        assert isinstance(rt_zeros, FastArray)

        assert_equal_(shape, rt_zeros.shape)
        assert_equal_(dtype.type, rt_zeros.dtype.type)
        assert_array_equal_(rt_zeros, np_zeros)

        # 1-D arrays always use Column-order. Otherwise, use the order specified
        if len(rt_zeros.shape) > 1 and min(rt_zeros.shape) > 1:
            assert np.isfortran(rt_zeros) == (order == "F")
        else:
            assert not np.isfortran(rt_zeros)

    @pytest.mark.xfail(reason="SOQTEST-6563 riptable does not implement subok")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    @given(
        arr=arrays(shape=ndarray_shape_strategy(), dtype=ints_floats_datetimes_and_timedeltas()),
        dtype=ints_floats_or_complex_dtypes(),
        order=sampled_from(("C", "F", "K", "A")),
        subok=booleans(),
    )
    def test_zeros_like(self, arr, dtype, order, subok):
        rt_zeros_like = rt.zeros_like(arr, dtype=dtype, order=order, subok=subok)
        np_zeros_like = np.zeros_like(arr, dtype=dtype, order=order, subok=subok)

        assert_equal_(dtype, rt_zeros_like.dtype)
        assert_equal_(rt_zeros_like.shape, arr.shape)
        assert_array_equal_(rt_zeros_like, np_zeros_like)

        # 1-D arrays always use Column-order. Otherwise, use the order specified
        if (
            len(rt_zeros_like.shape) > 1
            and min(rt_zeros_like.shape) > 1
            and (order == "F" or ((order == "A" or order == "K") and np.isfortran(arr)))
        ):
            assert np.isfortran(rt_zeros_like)
        else:
            assert not np.isfortran(rt_zeros_like)

        if subok:
            assert isinstance(rt_zeros_like, FastArray)
        else:
            assert isinstance(rt_zeros_like, np.ndarray)
            # FastArray is a subclass of np.ndarray, so also ensure it is not a FastArray
            assert not isinstance(rt_zeros_like, FastArray)



import pandas as pd
import pytest

from riptable import *

sorted_codes = FastArray(
    [
        2,
        4,
        4,
        3,
        2,
        1,
        3,
        2,
        0,
        1,
        3,
        4,
        2,
        0,
        4,
        3,
        1,
        0,
        1,
        2,
        3,
        1,
        4,
        2,
        2,
        3,
        4,
        2,
        0,
        2,
    ]
)
str_fa = FastArray(
    [
        "c",
        "e",
        "e",
        "d",
        "c",
        "b",
        "d",
        "c",
        "a",
        "b",
        "d",
        "e",
        "c",
        "a",
        "e",
        "d",
        "b",
        "a",
        "b",
        "c",
        "d",
        "b",
        "e",
        "c",
        "c",
        "d",
        "e",
        "c",
        "a",
        "c",
    ]
)
complete_unique_cats = FastArray(["a", "b", "c", "d", "e"])
big_cats = FastArray(["string" + str(i) for i in range(2000)])
matlab_codes = (sorted_codes + 1).astype(np.float32)


# TODO pytest parameterize expectations for dtypes
class TestCategoricalDtype:
    def test_codes(self):
        c = Categorical(sorted_codes.astype(np.int16), complete_unique_cats)
        assert c._fa.dtype == np.int16

        c = Categorical(sorted_codes.astype(np.uint64), complete_unique_cats)
        assert c._fa.dtype == np.int8

        c = Categorical(sorted_codes, complete_unique_cats, dtype=np.int64)
        assert c._fa.dtype == np.int64

        c = Categorical(sorted_codes, big_cats, dtype=np.int8)
        assert c._fa.dtype == np.int16

    def test_matlab(self):
        c = Categorical(matlab_codes, complete_unique_cats, from_matlab=True)
        assert c._fa.dtype == np.int8

        c = Categorical(matlab_codes, complete_unique_cats, from_matlab=True, dtype=np.int64)
        assert c._fa.dtype == np.int64

    def test_strings(self):
        c = Categorical(str_fa)
        assert c._fa.dtype == np.int8

        c = Categorical(str_fa, dtype=np.int64)
        assert c._fa.dtype == np.int64

    def test_strings_cats(self):
        c = Categorical(str_fa, complete_unique_cats)
        assert c._fa.dtype == np.int8

        c = Categorical(str_fa, complete_unique_cats, dtype=np.int64)
        assert c._fa.dtype == np.int64

    def test_multikey(self):
        c = Categorical([str_fa, sorted_codes])
        assert c._fa.dtype == np.int8

        c = Categorical([str_fa, sorted_codes], dtype=np.int64)
        assert c._fa.dtype == np.int64

    def test_pandas(self):
        pdc = pd.Categorical(str_fa)
        c = Categorical(str_fa)
        assert c._fa.dtype == np.int8

        pdc._codes = pdc._codes.astype(np.int32)
        c = Categorical(pdc)
        assert c._fa.dtype == np.int8

        c = Categorical(pdc, dtype=np.int64)
        assert c._fa.dtype == np.int64

    def test_errors(self):
        with pytest.raises(TypeError):
            c = Categorical(str_fa, dtype=np.uint8)

        with pytest.warns(UserWarning):
            c = Categorical(sorted_codes, big_cats, dtype=np.int8)



"""
Equality/comparison assertion functions for arrays, used when implementing riptable unit tests.
"""
__all__ = [
    "assert_array_or_cat_equal",
    "assert_categorical_equal",
    "assert_date_equal",
    "assert_datespan_equal",
    "assert_datetimenano_equal",
    "assert_timespan_equal",
]

from typing import Optional, TypeVar, Union

import numpy as np
from numpy.testing import assert_array_equal

from riptable.tests.test_utils import verbose_categorical

from .. import (
    Categorical,
    Date,
    DateScalar,
    DateSpan,
    DateSpanScalar,
    DateTimeNano,
    DateTimeNanoScalar,
    FastArray,
    TimeSpan,
    TimeSpanScalar,
)


def assert_categorical_equal(
    x: Categorical,
    y: Categorical,
    *,
    err_msg: Optional[str] = "",
    relaxed_check: bool = False,
    check_cat_names: bool = True,
    check_cat_types: bool = True,
    exact_dtype_match: bool = True,
) -> None:
    # TODO: Add optional flag parameters to customize the checks; e.g. we might want
    #       to have a flag controlling whether we require the category names to match,
    #       or the category dtypes to match, or the underlying dtype to match.
    # TODO: Would it make sense to have numpy assert_array_equal() be a @multidispatch function?
    #       Or even just a @singledispatch?
    #       Then we could just register this function as a handler for rt.Categorical.
    if not isinstance(x, Categorical):
        raise TypeError("'x' is not a Categorical.")
    elif not isinstance(y, Categorical):
        raise TypeError("'y' is not a Categorical.")

    extra_info = f"{x} {verbose_categorical(x)}\n\n{y} {verbose_categorical(y)}"

    # Require the Categoricals to have the same CategoryMode.
    if x.category_mode != y.category_mode:
        raise AssertionError(
            f"The Categoricals have different `CategoryMode` values: '{x.category_mode}' vs. '{y.category_mode}'.\n{extra_info}"
        )

    # Require the Categoricals to be the same shape (i.e., the shape of their underlying array data).
    if x._fa.shape != y._fa.shape:
        raise AssertionError(f"The Categoricals have different shapes: {x._fa.shape} vs. {y._fa.shape}.\n{extra_info}")

    # Do we have the same number of Categories in both?
    x_cat_dict = x.category_dict
    y_cat_dict = y.category_dict
    if len(x_cat_dict) != len(y_cat_dict):
        raise AssertionError(
            f"The Categoricals have different category arities (# of category columns): {len(x_cat_dict)} vs. {len(y_cat_dict)}.\n{extra_info}"
        )

    # Check category names match, in the same order.
    if check_cat_names:
        x_cat_names = list(x_cat_dict.keys())
        y_cat_names = list(y_cat_dict.keys())
        if x_cat_names != y_cat_names:
            raise AssertionError(
                f"The category name(s) are different between the two Categoricals:\t{x_cat_names} vs. {y_cat_names}.\n{extra_info}"
            )

    # Check the category types (Python types) match, in the same order.
    if check_cat_types:
        # The category array types must match exactly; if we only checked the dtypes (below) or that
        # one type was equal to or a subtype of the other, the semantic meaning of the data could be
        # lost. For example, consider if this array was a raw integer array in one Categorical and a
        # Date array in the other -- even if they contain the same underlying data, the Date type provides
        # some different interpretation/semantics on top of that data compared to a plain array.
        x_cat_types = [type(v).__qualname__ for v in x_cat_dict.values()]
        y_cat_types = [type(v).__qualname__ for v in y_cat_dict.values()]
        if x_cat_types != y_cat_types:
            x_cat_names_and_types = [(k, type(v).__qualname__) for k, v in x_cat_dict.items()]
            y_cat_names_and_types = [(k, type(v).__qualname__) for k, v in y_cat_dict.items()]
            raise AssertionError(
                f"The category array(s) have different types between the Categoricals:\t{x_cat_names_and_types} vs. {y_cat_names_and_types}.\n{extra_info}"
            )

    # If we're doing the relaxed check, just use .expand_array to expand the Categoricals to arrays
    # (or tuples of normal arrays) then check whether the pairs of arrays are equal.
    if relaxed_check:
        x_expanded_arrays = x.expand_dict
        y_expanded_arrays = y.expand_dict

        for i, ((x_cat_name, x_exp_arr), (y_cat_name, y_exp_arr)) in enumerate(
            zip(x_expanded_arrays.items(), y_expanded_arrays.items())
        ):
            # TODO: Use assert_array_or_cat_equal() here instead of assert_array_equal(), so that if the category arrays
            #       are a FastArray-derived array type, we'll use the type-specific equality checks which'll be stronger
            #       than generic array equality assertions.
            assert_array_equal(x_exp_arr, y_exp_arr, err_msg=err_msg + extra_info)
            isnan_kinds = "iuf"  # dtype 'kinds' for which FastArray supports .isnan()
            if np.dtype(x_exp_arr.dtype).kind in isnan_kinds:
                assert_array_equal(
                    x_exp_arr.isnan(),
                    y_exp_arr.isnan(),
                    err_msg=f"Different NaN/invalid values between the expanded arrays (x='{x_cat_name}', y='{y_cat_name}').\n{extra_info}",
                )

    else:
        # Check categories match.
        for i, ((x_cat_name, x_cat_arr), (y_cat_name, y_cat_arr)) in enumerate(
            zip(x_cat_dict.items(), y_cat_dict.items())
        ):
            # Check category dtypes match. If we're doing the relaxed form of this check,
            # only require that the dtypes have the same 'kind'.
            x_cat_arr_dtype = np.dtype(x_cat_arr.dtype)
            y_cat_arr_dtype = np.dtype(y_cat_arr.dtype)
            if exact_dtype_match:
                assert x_cat_arr_dtype == y_cat_arr_dtype, extra_info
            else:
                assert x_cat_arr_dtype.kind == y_cat_arr_dtype.kind, extra_info
                # TODO: we don't currently allow categoricals to be built out of recarrays / structured dtypes;
                #       but in case we ever do, we should add a check here so that if x_cat_arr_dtype.kind == 'V'
                #       we'll recursively apply this check to the dtype fields.

            # Strict check (for now) -- the category arrays *must* match.
            # We also need to use .isnan() on the arrays here to handle nan-equality (we want to consider nans equal
            # to each other for the purposes of this comparison).
            # TODO: Add a flag to relax this check, for when we just want to make sure that one set of categories
            #       is a subset of the other, and that every category in their intersection is assigned to the same
            #       value in both.
            # TODO: Use assert_array_or_cat_equal() here instead of assert_array_equal(), so that if the category arrays
            #       are a FastArray-derived array type, we'll use the type-specific equality checks which'll be stronger
            #       than generic array equality assertions.
            assert_array_equal(x_cat_arr, y_cat_arr, err_msg=err_msg + extra_info)
            isnan_kinds = "iuf"  # dtype 'kinds' for which FastArray supports .isnan()
            if x_cat_arr_dtype.kind in isnan_kinds:
                assert_array_equal(
                    x_cat_arr.isnan(),
                    y_cat_arr.isnan(),
                    err_msg=f"Different NaN/invalid values between the category arrays at index {i}.\n{extra_info}",
                )

        # The category array(s) match, check the underlying data matches too.
        assert_array_equal(x._fa, y._fa)

    # Check the Categoricals consider themselves to have invalids/NA values
    # in the same locations.
    assert_array_equal(
        x.isnan(), y.isnan(), err_msg="The Categorical arrays have invalid/NA values at different positions."
    )


def assert_date_equal(x: Union[Date, DateScalar], y: Union[Date, DateScalar], *, err_msg: Optional[str] = "") -> None:
    """Assert two `Date` arrays (and/or scalars) are equal."""
    # Both instances must be Date or DateScalar.
    if not isinstance(x, (Date, DateScalar)):
        raise TypeError(
            f"The 'x' parameter has type '{type(x).__qualname__}' but must be an instance of Date/DateScalar."
        )
    elif not isinstance(y, (Date, DateScalar)):
        raise TypeError(
            f"The 'y' parameter has type '{type(y).__qualname__}' but must be an instance of Date/DateScalar."
        )

    # Assert the underlying array data is equal.
    assert_array_equal(x._np, y._np, err_msg=err_msg)


def assert_datetimenano_equal(
    x: Union[DateTimeNano, DateTimeNanoScalar],
    y: Union[DateTimeNano, DateTimeNanoScalar],
    *,
    err_msg: Optional[str] = "",
) -> None:
    """Assert two `DateTimeNano` arrays (and/or scalars) are equal."""

    # Both instances must be DateTimeNano or DateTimeNanoScalar.
    if not isinstance(x, (DateTimeNano, DateTimeNanoScalar)):
        raise TypeError(
            f"The 'x' parameter has type '{type(x).__qualname__}' but must be an instance of DateTimeNano/DateTimeNanoScalar."
        )
    elif not isinstance(y, (DateTimeNano, DateTimeNanoScalar)):
        raise TypeError(
            f"The 'y' parameter has type '{type(y).__qualname__}' but must be an instance of DateTimeNano/DateTimeNanoScalar."
        )

    # Check the underlying array data is equal.
    # The ._np property works for both arrays and scalars.
    assert_array_equal(x._np, y._np, err_msg=err_msg)

    # Check the timezones match up.
    if x._timezone != y._timezone:
        raise AssertionError(
            f"The timezones are different: '{x._timezone._timezone_str}' vs. '{y._timezone._timezone_str}'."
        )


def assert_datespan_equal(
    x: Union[DateSpan, DateSpanScalar], y: Union[DateSpan, DateSpanScalar], *, err_msg: Optional[str] = ""
) -> None:
    """Assert two `DateSpan` arrays (and/or scalars) are equal."""
    # Both instances must be DateSpan or DateSpanScalar.
    if not isinstance(x, (DateSpan, DateSpanScalar)):
        raise TypeError(
            f"The 'x' parameter has type '{type(x).__qualname__}' but must be an instance of DateSpan/DateSpanScalar."
        )
    elif not isinstance(y, (DateSpan, DateSpanScalar)):
        raise TypeError(
            f"The 'y' parameter has type '{type(y).__qualname__}' but must be an instance of DateSpan/DateSpanScalar."
        )

    # Assert the underlying array data is equal.
    assert_array_equal(x._np, y._np, err_msg=err_msg)


def assert_timespan_equal(
    x: Union[TimeSpan, TimeSpanScalar], y: Union[TimeSpan, TimeSpanScalar], *, err_msg: Optional[str] = ""
) -> None:
    """Assert two `TimeSpan` arrays (and/or scalars) are equal."""
    # Both instances must be TimeSpan or TimeSpanScalar.
    if not isinstance(x, (TimeSpan, TimeSpanScalar)):
        raise TypeError(
            f"The 'x' parameter has type '{type(x).__qualname__}' but must be an instance of TimeSpan/TimeSpanScalar."
        )
    elif not isinstance(y, (TimeSpan, TimeSpanScalar)):
        raise TypeError(
            f"The 'y' parameter has type '{type(y).__qualname__}' but must be an instance of TimeSpan/TimeSpanScalar."
        )

    # Assert the underlying array data is equal.
    assert_array_equal(x._np, y._np, err_msg=err_msg)


_A = TypeVar("_A", bound=np.ndarray)


def assert_array_or_cat_equal(
    x: _A,
    y: _A,
    *,
    err_msg: Optional[str] = "",
    relaxed_cat_check: bool = False,
    check_cat_names: bool = True,
    check_cat_types: bool = True,
    exact_dtype_match: bool = True,
):
    # TODO: Add optional flag parameters to customize the checks; e.g. we might want
    #       to have a flag controlling whether we require the category names to match,
    #       or the category dtypes to match, or the underlying dtype to match.
    # TODO: Would it make sense to have numpy assert_array_equal() be a @multidispatch function?
    #       Or even just a @singledispatch?
    #       Then we could just register this function as a handler for rt.Categorical.
    # TODO: Add a check here for the name (if set) of the arrays themselves?
    #       E.g. if enabled, the check would verify both arrays are FastArray (or derived types)
    #       then call .get_name() on both and compare the results for equality.

    assert type(x) == type(y)

    # Categorical needs special handling, otherwise we run into TypeError in
    # Categorical.match_str_to_category when assert_array_equal() tries to index
    # it with an inf/-inf.
    if isinstance(x, Categorical):
        assert_categorical_equal(
            x,
            y,
            err_msg=err_msg,
            relaxed_check=relaxed_cat_check,
            check_cat_names=check_cat_names,
            check_cat_types=check_cat_types,
            exact_dtype_match=exact_dtype_match,
        )

    elif isinstance(x, (Date, DateScalar)):
        assert_date_equal(x, y, err_msg=err_msg)
    elif isinstance(x, (DateSpan, DateSpanScalar)):
        assert_datespan_equal(x, y, err_msg=err_msg)
    elif isinstance(x, (DateTimeNano, DateTimeNanoScalar)):
        assert_datetimenano_equal(x, y, err_msg=err_msg)
    elif isinstance(x, (TimeSpan, TimeSpanScalar)):
        assert_timespan_equal(x, y, err_msg=err_msg)

    elif isinstance(x, FastArray) and not type(x) == FastArray:
        # For FastArray-derived array types like Date, DateTimeNano, etc.
        # that don't otherwise have a handler defined above.
        # These don't work well with numpy's assert_array_equal when passed directly,
        # so we pass in views of the underlying numpy arrays.
        assert_array_equal(x._np, y._np, err_msg=err_msg)

        isnan_kinds = "iuf"  # dtype 'kinds' for which FastArray supports .isnan()
        x_dtype = np.dtype(x.dtype)
        if x_dtype.kind in isnan_kinds:
            assert_array_equal(x.isnan(), y.isnan(), err_msg=f"Different NaN/invalid values between the arrays.")

    else:
        # For other array types, we can just defer to numpy array assertions.
        # assert_allclose with equal_nan to compare NaN values
        assert_array_equal(x, y, err_msg=err_msg)

        isnan_kinds = "iuf"  # dtype 'kinds' for which FastArray supports .isnan()
        x_dtype = np.dtype(x.dtype)
        if x_dtype.kind in isnan_kinds:
            assert_array_equal(x.isnan(), y.isnan(), err_msg=f"Different NaN/invalid values between the arrays.")



from typing import Dict, List, Optional, Tuple, Union

import hypothesis.strategies as st
import numpy as np
from hypothesis import event
from hypothesis.extra.numpy import (
    array_shapes,
    arrays,
    byte_string_dtypes,
    integer_array_indices,
    integer_dtypes,
    unicode_string_dtypes,
    unsigned_integer_dtypes,
)

import riptable as rt
from riptable import Categorical
from riptable.rt_enum import CategoryMode

# TODO
#   * Implement a way to generate Categoricals of derived FastArray classes, such as rt.Date.
#     Could just implement strategies for each of them, then use .map(lambda x: rt.Cat(x)) when we need
#     a Categorical of that type.
#   * Implement a strategy which just creates a random Categorical of any type -- single/multikey, incl.
#     derived FastArray types. This could be our general go-to when we're implementing tests where we want
#     to ensure all Categorical cases are fully covered -- e.g. when a Categorical column in a Dataset is
#     round-tripped through SDS.
#   * Implement another strategy accepting one or more strategies and creates a single or multikey Categorical
#     with them. One difficulty with this _may_ be structuring this both so we allow for various strategies to be used
#     for the categorical data while also enforcing that each time we create an example, the array(s) produced
#     by the strategies must have the same length.
#   * Maybe implement (and contribute to hypothesis) a strategy like hypothesis.strategy.text but which generates
#     numpy arrays; this strategy should also have the unique= kwarg like 'arrays' does. If this strategy was available
#     it'd be useful as a building block for some of the Categorical strategies below -- it'd simplify some things
#     while (probably) also improving performance.


def category_labels(min_str_len: int, max_str_len: int, unicode: bool) -> st.SearchStrategy:
    """Creates a `hypothesis.strategies.SearchStrategy` instance for generating category labels."""
    category_label_chars = st.characters(
        whitelist_categories=("L", "Nd", "Nl", "Pc", "Pd", "Zs"),
        max_codepoint=None if unicode else 127,  # Max codepoint of 127 limits to ASCII codepoints.
    )
    return st.text(alphabet=category_label_chars, min_size=min_str_len, max_size=max_str_len)


@st.composite
def categorical_stringarray(
    draw,
    max_length: int,
    max_categories: int,
    *,
    endianness: str = "=",
    min_str_len: int = 1,
    max_str_len: int = 16,
    unicode: Optional[bool] = None,
    ordered: Optional[bool] = None,
) -> Categorical:
    """
    Strategy for creating StringArray-mode Categoricals.

    Parameters
    ----------
    draw
    max_length : int
    max_categories : int
    endianness : str
    min_str_len : int
    max_str_len : int
    unicode : bool, optional
    ordered : bool, optional

    Examples
    --------
    >>> array_strategy = arrays(integer_dtypes(endianness="=", sizes=(64,)), (5,))
    arrays(dtype=integer_dtypes(endianness='=', sizes=(64,)), shape=(5,))
    >>> categorical_stringarray(array_strategy, with_categories=True).example()
    0, 0, 0, 0, 0

    Notes
    -----
    TODO: Make sure to include the case where we have category values (in the underlying integer array)
          past the end of the categories array. (Or is that only for a Dictionary mode categorical?)
          To clarify -- this is the behavior where, when we print the Categorical, we get entries like <!456>.

    TODO: Also exercise (in one way or another) the following arguments to the Categorical constructor:
        * base_index
            Add an optional boolean parameter. When None, draw a boolean to fill it in.
            When the bool is false, call rt.Cat() with base_index=0.
            When True, call rt.Cat() with base_index=1.
        * dtype
            Call the ctor with dtype=None or a signed integer dtype that's either the min size given the
            number of categories or any larger signed integer dtype.
            E.g. if len(categories) == 1000, draw from { None, np.int16, np.int32, np.int64 }
        * filter
            Add an optional boolean param to the strategy which defaults to None, in which case we'll fill it by drawing a boolean.
            When the bool is false we we call rt.Cat() with filter=None.
            When True, we create a boolean array the same length as our values or fancy index and pass that as the filter.

    TODO: Support slicing/strides on the values/categories arrays passed to the Categorical constructor.

    TODO: When creating the fancy index array and we've drawn 'explicit_categories=True', allow the fancy index to be created
          with any applicable integer type (signed or unsigned) whose range is large enough to index into the categories array.
          (Or, should we just allow _any_ integer dtype, even if too small? We wouldn't be able to index categories past the
          range of the dtype, but maybe that's an interesting thing to test? Especially around cases like having auto_add=True.)
    """
    # Draw a boolean indicating how the data will be passed to the Categorical constructor later.
    # This is done first since it's one of the most likely things to affect the behavior of the Categorical,
    # and shrinking (in some cases) works better when such values are drawn earlier in strategy.
    explicit_categories: bool = draw(st.booleans())
    if explicit_categories:
        event("Categorical created from unique category array and fancy index.")
    else:
        event("Categorical created from non-unique array of strings.")

    # Draw the string dtype based on whether we want a byte (ascii) string or Unicode.
    is_unicode: bool = draw(st.booleans()) if unicode is None else unicode
    if is_unicode:
        labels_dtype = draw(unicode_string_dtypes(endianness=endianness, min_len=min_str_len, max_len=max_str_len))
    else:
        labels_dtype = draw(byte_string_dtypes(endianness=endianness, min_len=min_str_len, max_len=max_str_len))

    # Create an array of unique category labels.
    cats_shapes = array_shapes(max_dims=1, max_side=max_categories)
    category_label_strat = category_labels(min_str_len, max_str_len, unicode=is_unicode)
    unique_labels = draw(arrays(dtype=labels_dtype, shape=cats_shapes, elements=category_label_strat, unique=True))

    # Use basic_indices to create a fancy index into the array of unique category labels.
    # Apply it to expand the array of unique labels into an array where those labels may occur zero or more times.
    fancy_index_shapes = array_shapes(max_dims=1, max_side=max_length)
    fancy_index = draw(integer_array_indices(shape=unique_labels.shape, result_shape=fancy_index_shapes))

    # If the 'ordered' flag is not set, draw a boolean for it now so we have a concrete value
    # to use when creating the categorical.
    is_ordered = draw(st.booleans()) if ordered is None else ordered

    # If the 'explicit_categories' flag is set, create the Categorical by passing in the
    # unique values and fancy index separately.
    # Otherwise, apply the fancy index to the array of unique category values to produce an
    # array where each category appears zero or more times; then create the Categorical from that.
    if explicit_categories:
        return Categorical(fancy_index, categories=unique_labels, ordered=is_ordered, unicode=is_unicode)

    else:
        values = unique_labels[fancy_index]
        return Categorical(values, ordered=is_ordered, unicode=is_unicode)


def bimap(
    keys: st.SearchStrategy, values: st.SearchStrategy, *, min_size: int = 0, max_size: Optional[int] = None
) -> st.SearchStrategy:
    """
    Strategy similar to the hypothesis 'dictionaries' strategy, but which ensures both keys and values are unique.
    """
    return st.lists(st.tuples(keys, values), min_size=min_size, max_size=max_size, unique=True).map(dict)


@st.composite
def categorical_dictmode(
    draw,
    max_length: int,
    max_categories: int,
    *,
    endianness: str = "=",
    min_str_len: int = 1,
    max_str_len: int = 16,
    unicode: Optional[bool] = None,
    ordered: Optional[bool] = None,
) -> Categorical:
    """
    Strategy for creating Dictionary-mode Categoricals.

    This strategy currently only covers creating `Categorical` instances with
    string-typed category labels.

    Parameters
    ----------
    draw
    max_length : int
    max_categories : int
    endianness : str
    min_str_len : int
    max_str_len : int
    unicode : bool, optional
    ordered : bool, optional

    Examples
    --------
    >>> categorical_dictmode(10_000, 1_000, max_str_len=20).example()
    0, 0, 0, 0, 0

    Notes
    -----
    TODO: Make sure to include the case where we have category values (in the underlying integer array)
          past the end of the categories array. (Or is that only for a Dictionary mode categorical?)
          To clarify -- this is the behavior where, when we print the Categorical, we get entries like <!456>.

    TODO: Also exercise (in one way or another) the following arguments to the Categorical constructor:
        * base_index
            Add an optional boolean parameter. When None, draw a boolean to fill it in.
            When the bool is false, call rt.Cat() with base_index=0.
            When True, call rt.Cat() with base_index=1.
        * dtype
            Call the ctor with dtype=None or a signed integer dtype that's either the min size given the
            number of categories or any larger signed integer dtype.
            E.g. if len(categories) == 1000, draw from { None, np.int16, np.int32, np.int64 }
        * filter
            Add an optional boolean param to the strategy which defaults to None, in which case we'll fill it by drawing a boolean.
            When the bool is false we we call rt.Cat() with filter=None.
            When True, we create a boolean array the same length as our values or fancy index and pass that as the filter.

    TODO: Support slicing/strides on the values/categories arrays passed to the Categorical constructor.

    TODO: Does a Dictionary-mode Categorical allow any other types (e.g. rt.Date) to be used for the category labels?
        If so, these should also be covered by this strategy (though changes will needed to allow a variety of
        types to be used for category labels).

    TODO: Any possible issues (that we might want to exercise in this strategy) between the string used when displaying
        the invalid category (e.g. 'Inv') and category labels? What happens if we have a category label using the same string?
    """
    # Draw a boolean indicating whether we'll use a signed or unsigned integer dtype.
    use_signed_integer_dtype: bool = draw(st.booleans())

    # If using a signed integer dtype, draw another boolean indicating whether we'll
    # generate negative category values.
    allow_negative_category_values: bool = draw(st.booleans()) if use_signed_integer_dtype else False
    if use_signed_integer_dtype:
        if allow_negative_category_values:
            event("Categorical may have a mix of negative, zero, and positive category values.")
        else:
            event("Categorical has only non-negative category values.")

    # If the 'unicode' flag is not set, draw a boolean to fill it in.
    is_unicode: bool = draw(st.booleans()) if unicode is None else unicode
    event(f'Category labels are {"unicode" if is_unicode else "ascii"} strings.')

    # If the 'ordered' flag is not set, draw a boolean for it now so we have a concrete value
    # to use when creating the categorical.
    is_ordered = draw(st.booleans()) if ordered is None else ordered
    event(f"ordered = {is_ordered}")

    # Draw the dtype for the category values.
    # TODO: Draw a signed or unsigned integer dtype here which is at least as large as needed, but perhaps larger
    #       than needed.
    #       For now, we just use the smallest dtype large enough to fit the max number of categories; but allowing for
    #       larger (randomly-selected) dtypes later will help ensure we test cases where there are non-consecutive
    #       category values even when the max_categories value is near the max value of a dtype.
    values_dtype = np.min_scalar_type(max_categories)

    # Create the strategy for the category values (integer values representing the categories).
    values_dtype_info = np.iinfo(values_dtype)
    values_strat = st.integers(
        min_value=(values_dtype_info.min if allow_negative_category_values else 0), max_value=values_dtype_info.max
    )

    # Create an array of unique category values/codes.
    cats_shapes = array_shapes(max_dims=1, max_side=max_categories)
    unique_cat_values = draw(arrays(dtype=values_dtype, shape=cats_shapes, elements=values_strat, unique=True))

    # Draw the string dtype for the labels based on whether we want a byte (ascii) string or Unicode.
    is_unicode: bool = draw(st.booleans()) if unicode is None else unicode
    if is_unicode:
        labels_dtype = draw(unicode_string_dtypes(endianness=endianness, min_len=min_str_len, max_len=max_str_len))
    else:
        labels_dtype = draw(byte_string_dtypes(endianness=endianness, min_len=min_str_len, max_len=max_str_len))

    # Create an array of unique category labels; this must be the same shape as the unique category values array.
    category_label_strat = category_labels(min_str_len, max_str_len, unicode=is_unicode)
    unique_labels = draw(
        arrays(dtype=labels_dtype, shape=unique_cat_values.shape, elements=category_label_strat, unique=True)
    )

    # TODO: Draw a slice (or None) that we'll apply to both arrays of uniques (the labels and values)
    #   before using them to create the category dictionary.
    #   This allows us to cover cases where a category value isn't in the dictionary.

    # Combine the unique category labels and values to create a dictionary.
    category_dict = dict(zip(unique_labels, unique_cat_values))

    # Use basic_indices to create a fancy index into the array of unique values.
    # Apply it to expand the array of unique values into an array where those values may occur zero or more times.
    fancy_index_shapes = array_shapes(max_dims=1, max_side=max_length)
    fancy_index = draw(integer_array_indices(shape=unique_cat_values.shape, result_shape=fancy_index_shapes))

    # Apply the fancy index to the array of unique category values to produce an
    # array where each category appears zero or more times; then create the Categorical from that.
    cat_values = unique_cat_values[fancy_index]
    return Categorical(cat_values, categories=category_dict, ordered=is_ordered, unicode=is_unicode)


@st.composite
def categorical_intenum(draw) -> Categorical:
    """Strategy for creating IntEnum-mode Categoricals."""
    raise NotImplementedError()


@st.composite
def categorical_numericarray(draw) -> Categorical:
    """Strategy for creating NumericArray-mode Categoricals."""
    raise NotImplementedError()


class CategoricalStrategy(st.SearchStrategy):
    """
    A strategy for generating `Categoricals` across the `CategoryMode` \s using a
    hypothesis `SearchStrategy` for generating the values and categories.

    Parameters
    ----------
    value_strategy : SearchStrategy
        A strategy that generates :class:`numpy:numpy.ndarray` s or list-like values that
        the `Categorical` constructor accepts.
    with_categories : bool
        Whether to generate and pass categories when constructing the ``Categorical``.
    category_mode : ``CategoryMode``
        The type of `Categorical` to construct.
    ordered : bool
        Corresponds to ``Categorical`` ordered parameter when initializing.

    Examples
    --------
    >>> array_strategy = arrays(integer_dtypes(endianness="=", sizes=(64,)), (5,))
    arrays(dtype=integer_dtypes(endianness='=', sizes=(64,)), shape=(5,))
    >>> CategoricalStrategy(array_strategy, with_categories=True, category_mode=CategoryMode.StringArray).example()
    0, 0, 0, 0, 0
    """

    _CN = "CategoricalStrategy"

    def __init__(
        self,
        values_strategy: st.SearchStrategy,
        with_categories: bool = False,
        category_mode: Optional[CategoryMode] = None,
        ordered: Optional[bool] = None,
    ):
        super().__init__()
        self.value_strategy = values_strategy
        self.with_categories = with_categories
        self.category_mode = category_mode
        self.ordered = ordered

    def _construct_dict(self, data, values) -> Union[Dict[str, int], Dict[int, str]]:
        """Return a dictionary that will be used for Dictionary Categorymodes."""
        # add support for Dict[str, int] mode
        unique_values = np.unique(values)
        n = len(unique_values)
        keys = data.draw(
            st.lists(
                st.integers(min_value=1, max_value=n * 10),
                min_size=n,
                max_size=n,
                unique=True,
            )
        )
        return {k: str(v) for k, v in zip(keys, unique_values)}

    def do_draw(self, data):
        # categories will be set if either:
        # - with_categories parameter is set to True, or
        # - CategoryMode or category_mode designates a dictionary Categorical.
        values, categories, cat = None, None, None
        if self.category_mode == CategoryMode.StringArray:
            values = list(map(str, data.draw(self.value_strategy)))
            if self.with_categories:
                categories = list(map(str, set(values)))
            cat = Categorical(values, categories=categories, ordered=self.ordered)
        elif self.category_mode == CategoryMode.Dictionary:
            values = data.draw(self.value_strategy)
            category_dict = self._construct_dict(data, values)
            cat = Categorical(values, categories=category_dict, ordered=self.ordered)
        else:
            raise ValueError(f"{self._CN}.do_draw: unhandled category mode {self.category_mode}\n\t{self}")
        return cat

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f"{self._CN}({self.value_strategy}, with_categories={self.with_categories}, category_mode={self.category_mode} ordered={self.ordered})"


if __name__ == "__main__":
    from hypothesis.extra.numpy import arrays, integer_dtypes

    def print_cat(c):
        print(
            f"c\n{repr(c)}\n\ntype(c), mode\n{type(c)}, {rt.rt_enum.CategoryMode(c.category_mode)}\n"
            f"\nc._fa\n{repr(c._fa)}\n\nc.category_array\n{repr(c.category_array)}\n\nc.category_dict\n{c.category_dict}\n\n"
        )
        if c.category_mode == rt.rt_enum.CategoryMode.Dictionary:
            print(f"*c.category_mapping*\n{c.category_mapping}\n\n*c.category_codes*\n{c.category_codes}")

    cat_strat = categorical_dictmode(10_000, 1_000, max_str_len=20)
    print(f"cat_strat {cat_strat}\n")
    print_cat(cat_strat.example())



import random as rand

import numpy as np

functions_str = [
    # 'count',
    "sum",
    "mean",
    "median",
    "min",
    "max",
    # 'prod',
    "var",
    # 'quantile',
    # 'cumsum',
    # 'cumprod',
    # 'cummax',
    # 'cummin'
]

KEY_COLUMN_NAMES = "a b c d e g h i j k l m n o p r s t u v w x y z".split(" ")
VAL_COLUMN_NAMES = ["val" + char.upper() for char in KEY_COLUMN_NAMES]

VAL_COLUMNS_MAX = 8
VAL_COLUMNS_MIN = 1
VAL_COLUMNS_INC = 3

KEY_COLUMNS_MAX = 4
KEY_COLUMNS_MIN = 1
KEY_COLUMNS_INC = 1

SYMBOL_RATIO_MAX = 0.5
SYMBOL_RATIO_MIN = 0.1
SYMBOL_RATIO_INC = 0.2

AGGREGATION_MAX = 5
AGGREGATION_MIN = 1
AGGREGATION_INC = 1

SIZE_DEFAULT = 100  ## to ensure that multithreading is enabled


def safe_assert(self, ary1, ary2):
    epsilon = 0.00000005

    def isNaN(num):
        return num != num

    assert len(ary1) == len(ary2)

    for a, b in zip(ary1, ary2):
        assert abs(a - b) < epsilon or (isNaN(a) and isNaN(b))


class groupby_parameters:
    def random_agg_list(self, numb_funcs):
        lst = []
        while len(lst) < numb_funcs:
            val = rand.randint(0, len(functions_str) - 1)
            val = functions_str[val]

            if not val in lst:
                lst.append(val)

        return lst

    def __init__(self):
        self.val_cols = VAL_COLUMNS_MIN
        self.key_cols = KEY_COLUMNS_MIN
        self.symbol_ratio = SYMBOL_RATIO_MIN
        self.aggs = AGGREGATION_MIN
        self.agg_list = self.random_agg_list(self.aggs)  ##itertools.combinations(functions_str, PARAMETERS.aggs).

    def update(self, parameter=None):
        def update_aggs(self):
            self.aggs += AGGREGATION_INC
            value = self.aggs
            self.aggs %= AGGREGATION_MAX

            if self.aggs < AGGREGATION_MIN:
                self.aggs = AGGREGATION_MIN

            self.agg_list = self.random_agg_list(self.aggs)
            return value != self.aggs

        def update_symbs(self):
            self.symbol_ratio += SYMBOL_RATIO_INC
            value = self.symbol_ratio
            self.symbol_ratio %= SYMBOL_RATIO_MAX
            if self.symbol_ratio < SYMBOL_RATIO_MIN:
                self.symbol_ratio = SYMBOL_RATIO_MIN
            return value != self.symbol_ratio

        def update_val_cols(self):
            self.val_cols += VAL_COLUMNS_INC
            value = self.val_cols
            self.val_cols %= VAL_COLUMNS_MAX
            self.val_cols = self.val_cols if self.val_cols > VAL_COLUMNS_MIN else VAL_COLUMNS_MIN
            return value != self.val_cols

        def update_key_cols(self):
            self.key_cols += KEY_COLUMNS_INC
            value = self.key_cols
            self.key_cols %= KEY_COLUMNS_MAX
            self.key_cols = self.key_cols if self.key_cols > KEY_COLUMNS_MIN else KEY_COLUMNS_MIN
            return value != self.key_cols

        if parameter is None:
            if update_val_cols(self):
                if update_key_cols(self):
                    if update_symbs(self):
                        update_aggs(self)
        else:
            switch = {
                "aggs": update_aggs,
                "syms": update_symbs,
                "vals": update_val_cols,
                "keys": update_key_cols,
            }

            switch[parameter]()
            ###---------------------------------------------


###---------------------------------------------------------------------------


class groupby_everything:

    val_columns = VAL_COLUMNS_MIN
    key_columns = KEY_COLUMNS_MIN
    size = SIZE_DEFAULT
    symbol_ratio = SYMBOL_RATIO_MIN

    aggregation_functions_count = AGGREGATION_MIN
    symbol_ratio_count = int(
        (size * symbol_ratio) ** (1 / val_columns)
    )  ###this equation ensures an equal distribution in respect to numb columns
    aggregation_functions = []
    data = {}

    def init_dummy_data(self):
        for i in range(0, self.val_columns):
            self.data[VAL_COLUMN_NAMES[i]] = np.random.rand(self.size)

        for i in range(0, self.key_columns):
            self.data[KEY_COLUMN_NAMES[i]] = [
                "k" + str(rand.randint(0, self.symbol_ratio_count))
                for i in np.random.randint(self.size, size=self.size)
            ]

    def __init__(
        self,
        val_cols,
        key_cols,
        symbs,
        agg_list,
        sz=SIZE_DEFAULT,
    ):
        self.val_columns = val_cols  ##integer number of value columns
        self.key_columns = key_cols  ##integer number of key   columns
        self.size = sz  ###integer number of rows
        self.aggregation_functions = agg_list
        self.symbol_ratio = symbs  ###float   ratio of symbols to size
        self.symbol_ratio_count = int(
            (self.size * self.symbol_ratio) ** (1 / self.val_columns)
        )  ###number of expected groups

        self.data = {}  ###the groupby data
        self.init_dummy_data()
